open Logic
open Prenex
(*A weight is a type where the weight can be an integer value or infinity*)

type weight= Int of int | Infty

(*An indexprinc is a tuple of string and integer representing name of principal and its index*)

type indexprinc=string*int

(*A state is tuple of type weight and list of indexed principal*)

type state= State of (weight*indexprinc list)

(*A wt is a weighted tree that is a tuple of state and list of tuples containing a weight(representing weight of transitions) and wt(representing subtrees)*) 

type wt = WT of state * (weight * wt) list | Empty

let maxw = ref (Int 0);;


(*gen_princ generates a principal according to the values of the roles and maximum instance value*)
let gen_princ max r princlist= 
  (List.nth princlist (r-1),max);;

(* add i a function that adds a principal to a tree*)

let rec add i t =
  match t with
     WT(State(x,y),z)->
	let rec process l =
	  match l with 
	  (a,b)::xs->(a,add i b):: process xs
	  |[]->[]
	in
	  WT(State(x,i::y),process z)
    |Empty->Empty

(*tree takes as an input a role maximum instance value and a tree and  generates a list of tuples with weights of transitions and trees by adding appropriate principal to them *)

let maxw = ref (Int 0)

let rec tree r max princlist t =
  match r with 
    0->[]
    |_->
       let t1 = ((Int 0),add (gen_princ max r princlist) t) in 
	 t1::tree (r-1) max princlist t

(*gen_tree generates aweighted tree with all weights initialized to zero *)

let rec gen_tree m r plist=
 match m with
    0-> WT(State(Int 0,[]),[])
   |_->   let t = gen_tree (m-1) r plist in
	WT(State(Int 0,[]), tree r m plist t)

(*addw function add two weights*)
let addw x y =
 match x with 
    (Int x1)->
       (match y with 
	(Int y1)-> (Int (x1+y1))
	  |Infty->Infty)
   |Infty->
      (match y with 
	(Int y1)->Infty
	 |Infty->Infty)

(*process is a function to process alist of indexed principals and returns a list of only principal names*)

let rec processl pl =
 match pl with 
   (a,b)::pls->a::processl pls
   |[]->[]

(*maximum a list of transitions and extracts the maximum weight *)
let rec maximum maxw nlist =
match nlist with 
(Int(x1))::xs-> (match !maxw with
		   (Int(x2))->
		       if x1>x2 then (maxw:=Int(x1))
		       else (maxw:=Int(x2));maximum maxw xs
		   |Infty->(maxw:=Int(x1));maximum maxw xs)
|Infty::xs-> !maxw;maximum maxw xs 
|[]->maxw

(* the function max receives alist of tuples with first value being weight and other being a subtree and returns a list of weights only*)

let rec  max tlist =
 match tlist with
   (x,y)::xs->x::max xs
   |[]->[]

(* This defines two mutually recursive functions h_s and h_t that are heuristic functions that receive a tree and a formula and weights the states and transitions according to following sementics
             h_s(s,phi)= max (h_t(t,phi)) when princiapls can join
	  	       = -infinity when there are no join transitions and Phi= Forall... and the context doesnot contain an instance of A
		       = 0 otherwise
	     h_t(t,phi) = 1+ h_s (s',psi) when phi = Forall... and next state adds such principal
			  h_s(s',phi) when phi is = Forall .... and next states doesnot add such principal
			  1+ h_s (s',phi) when phi = Exists... and next state adds such principal
			  h_s(s',psi)  when phi is = Exists .... and next states doesnot add such principal
 *)

let rec 
h_t t phi = 
let formu = pnf phi in
  match formu with 
      Forall(x,y,psi)->
	(match t with
	   (s,WT(State(w,pl),nt))->
		 if (List.mem y (processl pl)) then
		  let (WT(State(a,b),c)) = h_s (WT(State(w,pl),nt)) psi in
		   let weigh =  (addw (Int 1) a) in
		     (weigh,WT(State(a,b),c))
		 else
		  let (WT(State(a,b),c))= h_s (WT(State(w,pl),nt)) phi in
		   let weigh= (addw (Int 0) a) in 
		     (weigh,(WT(State(a,b),c))) 
	   |(s,WT(State(w,pl),[]))->((Int 0),WT(State(w,pl),[])))
  
    |Exists(x,y,psi)-> 
       (match t with
	    (s,WT(State(w,pl),nt))->
	      if (List.mem y (processl pl)) then 
		let (WT(State(a,b),c)) = h_s (WT(State(w,pl),nt)) psi in
		let weigh = (addw (Int 0) a) in
		  (weigh,(WT(State(a,b),c))) 
	      else
		let (WT(State(a,b),c))= h_s (WT(State(w,pl),nt)) phi in
		let weigh= (addw (Int 1) a) in
		  (weigh,(WT(State(a,b),c))) 
	  |(s,WT(State(w,pl),[]))->((Int 0),WT(State(w,pl),[])))
    |_-> t

and h_s x theta =
maxw:=Int 0;
let formula = pnf theta in 
match x with
    WT(State(w,pl),[])->
      (match formula  with 
	   Forall(x,y,psi)->
	     if (List.mem y (processl pl)) then 
	       let weigh = (Int 0) in 
	       WT(State(weigh,pl),[])
	     else 
	       WT(State(Infty,pl),[])
	 |_->WT(State((Int 0),pl),[]))
  |WT(State(w,pl),z)->
     let rec proc z =
       match z with 
	 (n::ns)-> h_t n formula:: proc ns
	 |[]->[]
     in
     let wlist=proc z in
     let weigh = maximum maxw (max wlist) in 
       WT(State(!weigh,pl),wlist)

(*let tracelist= ref([]);;*)
(* the function trace_value will accept a list and a weight and will generate a list of pairs of which first element is the weight of whole trace ending on a leaf and that leaf is the second element of the pair *)
(* let rec trace_value wtree w = 
match wtree with 
  WT(State(a,b),[])-> let we = addw w a in tracelist:= (we,b)::!tracelist
  |WT(State(a,b),c)-> let we= addw w a in prol c we
and prol l w = 
match l with 
  (a,b)::[]->let we = addw w a in trace_value b we
  |(a,b)::xy->let we = addw w a in trace_value b we;prol xy w *)
 
 let rec trace_value wtree w = 
match wtree with
(WT(State(a,b),[])) -> let we = addw w a in 
			       (we,b)::[]
|(WT(State(a,b),c))-> let we = addw w a in prol c we
and prol l w =
match l with
|(a,b)::xy-> let we = addw w a in prol xy w @ trace_value b we
|[]->[];;

(*comparew function will compare two weights  and return true if the first is greater then other; false otherwise*)

let comparew a b  = 
match a,b with
(Int x),(Int y)-> if x>y then true else false
  |(Int x), Infty->true
  |Infty,(Int x)->false
  |Infty,Infty->false

(*sort_trace will sort the list in decreasing order of the weights putting pairs containing infty weights at the end of list*)

let rec sort_trace l = 
match l with 
(a,b)::ls-> let rec compare (x,y) nl =
    match nl with 
      (c,d)::nls->
	if (comparew x c) then  let temp  = (x,y) in  (compare temp nls)
	else let  temp=(c,d) in (compare temp nls)
      |[]->(x,y)
in let (w,z)=( compare (a,b) ls) in 
(w,z)::(sort_trace (List.filter (fun (u,v)-> if (w=u && z=v) then false else true) l))
  |[]->[]

(*sort_cc will sort a context (a list of principal instances) in alphabetical order this will be used to remove permutations of the contexts*)
let rec sort_ccc l =  match l with 
  (a,aa)::ls-> let rec sort_c  ll =
    	match ll with
        (cc,i)::ccs-> 
		let rec sort_cc (c,d) lll= 
  			match lll with
  	  		(y,ii)::zs-> 
  	     		 if (c < y) then let temp = (c,d) in (sort_cc temp zs)
  	     	 	else let  temp = (y,ii) in (sort_cc temp zs)
  	  		|[]-> (c,d)
   		in
	 	let (u,v)=(sort_cc (cc,i) ccs )
		in
  	(u,v)::(sort_c (List.filter (fun (w,z) -> if (u=w && v=z) then false else true) ll))
	|[]->[]
 	in
  	(a,sort_c aa)::sort_ccc ls
  |[]->[]


(*max_perm will remove the indices from the contexts this will alllow to remove the permutation of contexts easily*)
let rec max_perm l = match l with 
(a,b)::ls-> let rec procl p = match p with 
(w,z)::ps->w::procl ps
  |[]->[] in
(a,procl b)::max_perm ls
  |[]->[]

(*remove the contexts with same principals with different index numnbers*)
let rec rem_per l = match l with
(x,y)::xs->(x,y)::rem_per (List.filter (fun (u,v) -> if v = y then false else true) l)
  |[]->[]

let ntree t pl = 
let htree=h_s t (pnf pl) in htree
