(*
	utils.ml
	01/05/2003
	Giacomo Baldi

	Implementation of utility functions
*)


(*list functions*)

exception UtilException

let rec len =
  function l ->
    match l with
    | x::xs -> 1 + (len xs)
    | [] -> 0
	  
let rec select =
  function l ->
    function n ->
      match l with
      | x::xs when n=0 -> x
      | x::xs -> select (xs) (n-1)
      | [] -> print_string("select!\n"); raise UtilException

let rec debox =
  function l ->
    match l with
    | x::xs -> x@(debox xs)
    | [] -> []

let rec map =
  function l ->
    function f ->
      match l with
      | x::xs -> (f x)::(map xs f)
      | [] -> []

let rec iter =
  function l ->
    function f ->
      match l with
      | x::xs -> let y = (f x) in  (iter xs f)
      | [] -> ()


let rec filter =
  function l ->
    function p ->
      match l with
      | x::xs -> if (p x) then x::(filter xs p) else (filter xs p)
      | [] -> []

let rec mul =
  function la ->
    function lb ->
      function f ->
	match la with
	| a::aa -> (map lb (f a))@(mul aa lb f)
	| [] -> []

let rec exists =
  function l ->
    function p ->
      match l with
      | x::xs -> if (p x) then true else exists xs p
      | [] -> false

let rec forall =
  function l ->
    function p ->
      match l with
      | x::xs -> if (p x) then (forall xs p) else false
      | [] -> true

let rec remclone =
  function l ->
    function p ->
      match l with
      | x::xs -> x::(remclone (filter xs (p x)) p)
      | [] -> []

let diff =
  function l1 ->
    function l2 ->
      filter l1 (function x -> forall l2 (function y -> y<>x ))


let rec print_list =
  function ll ->
    function pr ->
      match ll with
      | x::xs -> pr x; print_list xs pr
      | [] -> print_string("\n")



let rec fold =
  function l ->
    function f ->
      function en ->
	match l with
	| ll::[] -> f ll en
	| ll::ls -> f ll (fold ls f en)
	| [] ->  print_string("fold!\n");raise UtilException


let rec fold2 =
  function l ->
    function f ->
      match l with
      | ll1::ll2::[] -> f ll1 ll2
      | ll::[] ->  print_string("fold2_1!\n"); raise UtilException
      | [] ->  print_string("fold2_2!\n");raise UtilException
      | ll::ls -> f ll (fold2 ls f )


let head =
  function l ->
    match l with
    | ll::ls -> ll
    | [] ->  print_string("head!\n");raise UtilException

let tail =
  function l ->
    match l with
    | ll::ls -> ls
    | [] -> []



let rec aleph =
  function l ->
    match l with
    | ll::ls ->
	(let ss = aleph ls in
	(mul [ll] (aleph ls) (function a -> function b -> a::b))@ss
	)
    | [] -> [[]]


	  
let rec rebuild =
  function l1 ->
    function l2 ->
      function f ->
	match (l1,l2) with
	| x::xs,y::ys -> (mul ([x]) y f)@(rebuild xs ys f)
	| [],[] -> []



let rec reverse =
  function l ->
    match l with
    | x::xs -> (reverse xs)@[x]
    | [] -> []


let rec grow =
  function a ->
    function n ->
      if n<=0 then [] else a::(grow a (n-1))



let rec couplify =
  function l1 ->
    function l2 ->
      match l1,l2 with
      | x::xs, y::ys -> (x,y)::(couplify xs ys)
      | [],[] -> []
      | _ -> print_string("Couplify!\n"); raise UtilException


let rec tensor =
  function ll ->
    function f ->
      match ll with
      | x::xs -> 
	  ( (f x)::xs)::(mul [x] (tensor xs f) ( function a -> function b -> a::b))
      | [] -> []

let rec partition =
  function elements ->
    let rec nextpart =
      function part ->
	function e ->
	  match part with
	  | y::[] -> ([[e]]@y)::(tensor y (function a -> e::a))
	  | y::ys ->
let newpart = ([[e]]@y)::(tensor y (function a -> e::a)) in
	      newpart@(nextpart ys e)
	  | [] -> [[[e]]]
    in
    match elements with
    | x::xs -> 
	let prevp = partition xs in
	nextpart prevp x
    | [] -> []
		    





let compose =
  function f1 ->
    function f2 ->
      function x ->
	f2 (f1 x)

let invert =
  function f ->
    (
     function x ->
       function y -> f y x
    )

let half =
  function f ->
    (
     function x -> f x x
    )



let rec sort =
function l ->
function lt ->
if l=[] then l else
(
 let top = head l in
 let half1 = filter l (lt top) in
 let half2 = filter l (compose (lt top) (not)) in
 if (half1=[] || half2=[]) then l 
 else (sort half1 lt)@(sort half2 lt)
)
