(*
	Assignment.mli
	23/01/2003
	Giacomo Baldi
	
	Definition of assignment between variable names and values (messages)
*)
(**
	This module implements substitution application. A substitution is a set of couples (v,m) where
	v is a variable (open or symbolic) and m is the message associated 	with v (v -> m).
*)


open Message

(**
	raised when trying to substitute a malformed message.	
*)
exception NoMeaning

(**
	raised when trying to add a non-variable to the assignment.
*)
exception NeedVariable

(**
	an assignment is a function that given a message m, produces a message n, applying
	the substitution of each variable v with the associated message in the substitution table.
*)
type assignment = msg -> msg

(**
	void takes no parameters.
	@return an assignment with an empty substitution
*)
val void: unit -> assignment

(**
	isVoid takes an assignment a.
	@return true if the substitution associated with a is empty.
*)
val isVoid: assignment -> bool

(**
	add takes an assignment a, a message v and a message m.
	@return an assignment with the same subtitution of a plus the associtation v -> m
	@raise NeedVariable if v isn't a variable.
*)
val add: assignment -> msg -> msg -> assignment

(**
	* takes an assignment alfa and an assignment beta.
	@return the composition of alfa and beta.
	a composition of two assignment a and b, is a new assignment with
	the associtated substitution which is the transitive closure of the substitution of
	a and b.
*)
val ( * ): assignment -> assignment -> assignment



(**
   leftof takes an assignment a and a variable (Var) v.
   @return the list of variables whose right part in assignment a is v
*)
val leftof: assignment -> msg ->msg list	      

(**
	prnt takes an assignment a.
	Prints out a representation of the substitution of a.
*)
val prnt: assignment -> unit

(**
	prntHtml takes an assignment a.
	Prints out a representation of the substitution of a.
*)
val prntHtml: assignment -> unit

(**
	tostring takes an assignment a.
	@return a string representation of the substitution of a.
*)
val tostring: assignment -> string

(**
	comp takes an assignment alfa and an assignment beta.
	@return the composition of alfa and beta.
	a composition of two assignment a and b, is a new assignment with
	the associtated substitution which is the transitive closure of the substitution of
	a and b.
*)
val comp: assignment -> assignment -> assignment

(**
	listfy takes an assignment a.
	@return a list representation of the subtitution of a.
*)
val listfy: assignment -> (msg*msg) list

(**
	newasg takes a substitution s.
	@return an assignment associated with s.
*)
val newasg: (msg*msg) list -> assignment
