
(*
	core.ml
	26/07/2003
	Giacomo Baldi

	Implementation of core functions: l, llazy, mu, ni

*)

open Message
open Knowledge
open Kmanager
open Assignment
open Utils

exception Undefined




(* Symbolic functions *)
let rec l =
  function kman ->
    function asg ->
      function mm ->
	match mm with
	| K(_)
	| PNp(_,_)
	| PNm(_,_)
	| NO(_,_)
	| PN(_,_) -> [Assignment.void()]
	| SVar(v,i,_) -> [Assignment.void()]
	| BVar(v) -> [Assignment.void()]

	| Publ(SVar(v,i,RName))
	| Priv(SVar(v,i,RName)) -> [Assignment.void()]
	| Publ(SVar(v,i,Name))
	| Priv(SVar(v,i,Name)) ->
	    let kr = if (Knowledge.hasRNames (Kmanager.getK kman i)) then [(SVar(v,i,RName))] else [] in
	    let kk = (Knowledge.getNames (Kmanager.getK kman i))@kr in
	    let res1 = map kk (function pn -> Assignment.add (Assignment.void())  (SVar(v,i,Name)) (pn)) in
	    res1
	| Publ(SVar(v,i,t)) when (t=Gen || t=GenR) ->
	    let res1 = [Assignment.add (Assignment.void()) (SVar(v,i,t)) (SVar(v,i,Name))] in
	    let res2 = (map res1 (function sigma -> l kman (asg) (sigma mm))) in
	    rebuild res1 res2 (function (s1) -> function (s2) -> (s2*s1))
	| Priv(SVar(v,i,t)) when (t=Gen || t=GenR) ->
	    let res1 = [Assignment.add (Assignment.void()) (SVar(v,i,t)) (SVar(v,i,Name))] in
	    let res2 = (map res1 (function sigma -> l kman (asg) (sigma mm))) in
	    rebuild res1 res2 (function (s1) -> function (s2) -> (s2*s1))

	| Couple(m1,m2) ->
	    let res1 = (l kman asg m1) in
	    let res2 = (map res1 (function sigma -> l kman (asg) (sigma m2))) in
	    rebuild res1 res2 (function (s1) -> function (s2) -> (s2*s1))

	| Crypt(m, Publ(SVar(v,k,Name)))
	| Crypt(m, Priv(SVar(v,k,Name))) ->
	    let kr = if (Knowledge.hasRNames (Kmanager.getK kman k)) then [(SVar(v,k,RName))] else [] in
	    let kk = (Knowledge.getNames (Kmanager.getK kman k))@kr in
	    let res1 = map kk (function pn -> Assignment.add (Assignment.void())  (SVar(v,k,Name)) (pn)) in
	    let res2 = (map res1 (function sigma -> l kman (asg) (sigma m))) in
	    rebuild res1 res2 (function (s1) -> function (s2) -> (s2*s1))

	| Crypt(m, Publ(SVar(v,k,RName)))
	| Crypt(m, Priv(SVar(v,k,RName))) ->
	    l kman asg m

	| Crypt(m, SVar(v,k,t)) when (Knowledge.hasKeys (Kmanager.getK kman k)) && (t=Gen || t=GenR) ->
	    let oldk = (Kmanager.getK kman k) in
	    let ks = if (Knowledge.hasSKeys oldk) then [SVar(v,k,SKey)] else [] in
	    let kpm = if (Knowledge.hasPMKeys oldk) then [SVar(v,k,PMKey)] else [] in
	    let kmp = if (kpm<>[]) then [SVar(v,k,MPKey)] else [] in
	    let kps = Knowledge.getPKeys oldk in
	    let kms = Knowledge.getMKeys oldk in
	    let kk = ks@kpm@kmp@kps@kms in
	    let res1 = map kk (function pn -> Assignment.add (Assignment.void())  (SVar(v,k,t)) (pn)) in
	    let res2 = (map res1 (function sigma -> l kman (asg) (sigma m))) in
	    rebuild res1 res2 (function (s1) -> function (s2) -> (s2*s1))

	| Crypt(m, Priv(SVar(v,k,t)))
	| Crypt(m, Publ(SVar(v,k,t))) when (t=Gen || t=GenR) ->
	    let sr = Assignment.add (Assignment.void()) (SVar(v,k,t)) (SVar(v,k,Name)) in
	    let res = l kman asg (sr mm) in
	    map res (function sigma -> sigma*sr)

	| Crypt(m, SVar(v,k,t)) -> l kman asg m
	| Crypt(m, kk) when isGroundKey kk -> l kman asg m
	| Crypt(m,Publ(_)) -> l kman asg m
	| Crypt(m,Priv(_)) -> l kman asg m
	| Crypt(m, _) -> l kman asg m
	| _ -> []
and mu =
  function kman ->
    function k ->
      function asg ->
	function mm -> (
	  match mm with
	  | BVar(v) when not (Knowledge.isEmpty k) -> [(Var(v), Assignment.add (Assignment.void()) (Var(v)) (SVar(v,Knowledge.id_of(k),Gen)))]
	  | K(_) when (Knowledge.derives k (asg) mm) ->  [(mm, Assignment.void())]
	  | PN(_,_) when (Knowledge.derives k (asg) mm) ->  [(mm, Assignment.void())]
	  | PNp(_,_) when (Knowledge.derives k (asg) mm) ->  [(mm, Assignment.void())]
	  | PNm(_,_) when (Knowledge.derives k (asg) mm) ->  [(mm, Assignment.void())]
	  | NO(_,_) when (Knowledge.derives k (asg) mm) ->  [(mm, Assignment.void())]

	  | SVar(v,i,Gen) -> [(SVar(v,i,GenR), Assignment.add (Assignment.void()) (SVar(v,i,Gen)) (SVar(v,i,GenR)))]
	  | SVar(v,i,t) -> [(SVar(v,i,t), (Assignment.void()))]

	  | Publ(SVar(v,i,RName)) -> [(mm, (Assignment.void()))]
	  | Priv(SVar(v,i,RName)) -> [(mm, (Assignment.void()))]

	  | Couple(m1, m2) ->
	      let res1 = mu kman k asg m1 in

(*
   print_string("fst\n");
   prntMsg m1;
   print_string("\nres1: ");
   print_list res1 (function (m1,s1) -> Assignment.prnt s1);
   print_string("\n");
 *)
	      let resl = debox (map res1 (function (m1,s1) -> let tt = l kman s1 (s1 m2) in map tt (function s2 -> (s2 m1, s2*s1)))) in
	      let res2 = (map (map resl (snd)) (function sigma -> mu kman k (asg) (sigma m2))) in
	      (rebuild resl res2 (function (m1,s1) -> function (m2,s2) -> (Couple(m1,m2), s1*s2)))
		
		

	  | Crypt(m, K(s,i)) -> (
	      let res =
		if (Knowledge.derives k (asg) (K(s,i))) then map (mu kman k asg m) (function (m1,s1) -> (Crypt(m1, K(s,i)), s1))
		else []
	      in
	      let kl = (Knowledge.getCrypts k (K(s,i))) in
	      res@(mulist kman k asg kl m)
	     )
	  | Crypt(m, PNp(s,p)) ->  (
	      let res =
		if (Knowledge.derives k (asg) (PNm(s,p))) then map (mu kman k asg m) (function (m1,s1) -> (Crypt(m1, PNm(s,p) ), s1))
		else []
	      in
	      let kl = (Knowledge.getCrypts k (PNm(s,p))) in
	      res@(mulist kman k asg kl m)
	     )
	  | Crypt(m, PNm(s,p)) -> (
	      let res =
		if (Knowledge.derives k (asg) (PNp(s,p))) then map (mu kman k asg m) (function (m1,s1) -> (Crypt(m1, PNp(s,p)), s1))
		else []
	      in
	      let kl = (Knowledge.getCrypts k (PNp(s,p))) in
	      res@(mulist kman k asg kl m)
	     )

	  | Crypt(m, SVar(v,i,SKey)) ->
	      let res = mu kman k asg m in
	      map res (function (m1,s1) -> (Crypt(m1, (SVar(v,i,SKey))),s1))
(*
   let oldk = Kmanager.getK kman i in
   let res =
   if (Knowledge.hasSKeys oldk) then map (mu kman k asg m) (function (m1,s1) -> (Crypt(m1,SVar(v,i,SKey)), s1))
   else []
   in res
 *)
(*				let kl = Knowledge.getSCrypts oldk asg in
   let res2 = map (mulist kman k asg kl m) (function (Crypt(m1,kk),s1) -> (Crypt(m1,kk),s1*(Assignment.add (Assignment.void()) (SVar(v,i,SKey)) (kk)))) in
   res@res2
 *)
	  | Crypt(m, Publ(SVar(v,i,RName))) ->
	      let res = mu kman k asg m in
	      map res (function (m1,s1) -> (Crypt(m1, Priv(SVar(v,i,RName))),s1))
	  | Crypt(m, Priv(SVar(v,i,RName))) ->
	      let res = mu kman k asg m in
	      map res (function (m1,s1) -> (Crypt(m1, Publ(SVar(v,i,RName))),s1))

	  | Crypt(m, SVar(v,i,PMKey)) ->
	      let res = mu kman k asg m in
	      map res (function (m1,s1) -> (Crypt(m1, SVar(v,i,MPKey)),s1))

	  | Crypt(m, SVar(v,i,MPKey)) ->
	      let res = mu kman k asg m in
	      map res (function (m1,s1) -> (Crypt(m1, SVar(v,i,PMKey)),s1))

	  | Crypt(m, SVar(v,i,t)) ->
	      let res = mu kman k asg m in
	      map res (function (m1,s1) -> (Crypt(m1, SVar(v,i,t)),s1))
	  | _ -> [] (*map (l kman asg mm) (function s -> (s mm,s))*)
	 )
and mulist =
  function kman ->
    function k ->
      function asg ->
	function cc ->
	  function n ->
	    match cc with
	    | Crypt(m,kk)::ccs ->
		let rr = ni kman asg m n in
		(map rr (function r -> (Crypt(m,kk),r)))@(mulist kman k asg ccs n)
	    | [] -> []
and ni =
  function kman ->
    function asg ->
      function m ->
	function n ->
	  match m,n with
	  | _ , BVar(v) -> [Assignment.add (Assignment.void()) (Var(v)) m]
	  | K(_), _ when n=m ->  [(Assignment.void())]
	  | PN(_,_), _  when n=m ->  [(Assignment.void())]
	  | PNp(_,_), _ when n=m ->  [(Assignment.void())]
	  | PNm(_,_), _ when n=m ->  [(Assignment.void())]
	  | NO(_,_), _ when n=m ->  [(Assignment.void())]
	  | Crypt(mm,K(s,i)), Crypt(nn,K(w,y)) when (s=w && i=y)-> ni kman asg mm nn
	  | Crypt(mm,PNp(s1,p1)), Crypt(nn,PNm(s2,p2)) when (s1=s2 && p1=p2) -> ni kman asg mm nn
	  | Crypt(mm,PNm(s1,p1)), Crypt(nn,PNp(s2,p2)) when (s1=s2 && p1=p2) -> ni kman asg mm nn
	  | Crypt(mm, SVar(v1,i,SKey)), Crypt(nn, SVar(v2,j,SKey)) ->
	      let min = if i<j then i else j in
	      let ss = [Assignment.add (Assignment.add (Assignment.void()) (SVar(v1,i,SKey)) (SVar(v1,min,SKey))) (SVar(v2,j,SKey)) (SVar(v1,min,SKey))] in
	      let res = map ss (function sigma -> ni kman asg (sigma mm) (sigma nn)) in
	      rebuild ss res (function (s1) -> function (s2) -> (s1*s2))

	  | Couple(mm1,mm2), Couple(nn1,nn2) ->
	      let res1 = ni kman asg mm1 nn1 in
	      let res2 = (map res1 (function sigma -> ni  kman (asg*sigma) (sigma mm2) (sigma nn2))) in
	      rebuild res1 res2 (function (s1) -> function (s2) -> (s1*s2))
	  | SVar(v1,i,Gen), SVar(v2,j,Gen) ->
	      let min = if i<j then i else j in
	      [Assignment.add (Assignment.add (Assignment.void()) m (SVar(v1,min,Gen))) n (SVar(v1,min,GenR))]

	  | SVar(v1,i,t1), SVar(v2,j,t2) when comparable(t1,t2) ->
	      let min = if i<j then i else j in
	      [Assignment.add (Assignment.add (Assignment.void()) m (SVar(v1,min,mintype(t1,t2)))) n (SVar(v1,min,mintype(t1,t2)))]

	  | Publ(SVar(v1,i,t1)), Publ(SVar(v2,j,t2)) when comparable(t1,t2) ->
	      let min = if i<j then i else j in
	      [Assignment.add (Assignment.add (Assignment.void()) (SVar(v1,i,t1)) (SVar(v1,min,mintype(t1,t2)))) (SVar(v2,j,t2)) (SVar(v1,min,mintype(t1,t2)))]
	  | Priv(SVar(v1,i,t1)), Priv(SVar(v2,j,t2)) when comparable(t1,t2) ->
	      let min = if i<j then i else j in
	      [Assignment.add (Assignment.add (Assignment.void()) (SVar(v1,i,t1)) (SVar(v1,min,mintype(t1,t2)))) (SVar(v2,j,t2)) (SVar(v1,min,mintype(t1,t2)))]
	  | SVar(v1,i,t1), Publ(SVar(v2,j,t2)) when (t1=Gen || t1=GenR)&&(t2=Name || t2=RName) ->
	      let min = if i<j then i else j in
	      [Assignment.add (Assignment.add (Assignment.void()) m (Publ(SVar(v1,min,t2)))) (SVar(v2,j,t2)) ((SVar(v2,min,t2)))]
	  | SVar(v1,i,t1), Priv(SVar(v2,j,t2)) when (t1=Gen || t1=GenR)&&(t2=Name || t2=RName) ->
	      let min = if i<j then i else j in
	      [Assignment.add (Assignment.add (Assignment.void()) m (Priv(SVar(v1,min,t2)))) (SVar(v2,j,t2)) ((SVar(v2,min,t2)))]
	  | Publ(SVar(v2,j,t2)), SVar(v1,i,t1)  when (t1=Gen || t1=GenR)&&(t2=Name || t2=RName) ->
	      let min = if i<j then i else j in
	      [Assignment.add (Assignment.add (Assignment.void()) (SVar(v2,j,t2)) ((SVar(v1,min,t2)))) n (Publ((SVar(v2,min,t2))))]
	  | Priv(SVar(v2,j,t2)), SVar(v1,i,t1)  when (t1=Gen || t1=GenR)&&(t2=Name || t2=RName) ->
	      let min = if i<j then i else j in
	      [Assignment.add (Assignment.add (Assignment.void()) (SVar(v2,j,t2)) ((SVar(v1,min,t2)))) n (Priv((SVar(v2,min,t2))))]
	  | sm, SVar(v,i,Gen)  when notSVar(sm) ->
	      let res = mu kman (Kmanager.getK kman i) asg (m) in
	      map res (function (mm,ss) -> (ss*(Assignment.add (Assignment.void()) n mm)))
	  | sm, SVar(v,i,GenR)  when notSVar(sm) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu kman oldk asg (m) in
	      let res2 = filter res (function (mm,ss) -> Knowledge.knows oldk mm) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) n mm))
	  | sm, SVar(v,i,SKey)  when notSVar(sm) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu kman oldk asg (m) in
	      let res2 = filter res (function (mm,ss) -> ((isskey mm))) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) n mm))
	  | sm, SVar(v,i,Name)  when notSVar(sm) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu kman oldk asg (m) in
	      let res2 = filter res (function (mm,ss) -> ((isname mm))) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) n mm))
	  | sm, SVar(v,i,RName)  when notSVar(sm) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu kman oldk asg (m) in
	      let res2 = filter res (function (mm,ss) -> (Knowledge.isRName oldk mm )) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) n mm))
	  | sm, SVar(v,i,t)  when (notSVar(sm) && (t=MPKey ||  t=PMKey)) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu kman oldk asg (m) in
	      let res2 = filter res (function (mm,ss) -> Knowledge.isPMKey oldk mm) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) n mm))

	  | SVar(v,i,Gen), sm  when notSVar(sm) ->
	      let res = mu kman (Kmanager.getK kman i) asg (n) in
	      map res (function (mm,ss) -> (ss*(Assignment.add (Assignment.void()) m mm)))
	  | SVar(v,i,GenR), sm  when notSVar(sm) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu kman oldk asg (n) in
	      let res2 = filter res (function (mm,ss) -> Knowledge.knows oldk mm) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) m mm))
	  | SVar(v,i,SKey), sm  when notSVar(sm) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu kman oldk asg (n) in
	      let res2 = filter res (function (mm,ss) -> ((isskey mm))) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) m mm))
	  | SVar(v,i,Name), sm  when notSVar(sm) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu kman oldk asg (n) in
	      let res2 = filter res (function (mm,ss) -> ((isname mm))) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) m mm))
	  | SVar(v,i,RName), sm  when notSVar(sm) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu kman oldk asg (n) in
	      let res2 = filter res (function (mm,ss) -> (Knowledge.isRName oldk mm )) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) m mm))
	  | SVar(v,i,t), sm  when (notSVar(sm) && (t=MPKey ||  t=PMKey)) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu kman oldk asg (n) in
	      let res2 = filter res (function (mm,ss) -> Knowledge.isPMKey oldk mm) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) m mm))
	  | SVar(v,i,_), sm when notSVar(sm) ->
	      let res = mu kman  (Kmanager.getK kman i) asg (n) in
	      map res (function (mm,ss) -> (ss*(Assignment.add (Assignment.void()) m mm)))
	  | _ -> []



(* unification functions *)
let rec mu_eq =
  function kman ->
    function k ->
      function asg ->
	function mm -> (
	  match mm with
	  | Var(v) -> if (Knowledge.hasSKeys (Kmanager.getK kman  1)) then [( SVar(v,1,SKey),  (Assignment.add (Assignment.void()) mm (SVar(v,1,SKey)) ))] else []
	  | K(_) when (Knowledge.derives k (asg) mm) ->  [(mm, Assignment.void())]
	  | PN(_,_) when (Knowledge.derives k (asg) mm) ->  [(mm, Assignment.void())]
	  | PNp(_,_) when (Knowledge.derives k (asg) mm) ->  [(mm, Assignment.void())]
	  | PNm(_,_) when (Knowledge.derives k (asg) mm) ->  [(mm, Assignment.void())]
	  | NO(_,_) when (Knowledge.derives k (asg) mm) ->  [(mm, Assignment.void())]
	  | SVar(v,i,Gen) -> [(mm, Assignment.void())]
	  | SVar(v,i,GenR) -> [(mm, Assignment.void())]

	  | SVar(v,i,Name) ->
	      let oldk = Kmanager.getK kman i in
	      let things = Knowledge.getAllNames oldk in
	      let res = map things (function x -> (x, Assignment.add (Assignment.void()) (mm) x)) in
	      res
	  | SVar(v,i,RName) ->
	      let oldk = Kmanager.getK kman i in
	      let things = Knowledge.getRNames oldk in
	      let res = map things (function x -> (x, Assignment.add (Assignment.void()) (mm) x)) in
	      res
	  | SVar(v,i,SKey) ->
	      let oldk = Kmanager.getK kman i in
	      let things = Knowledge.getSKeys oldk in
	      let res = map things (function x -> (x, Assignment.add (Assignment.void()) (mm) x)) in
	      res
	  | SVar(v,i,MPKey) ->
	      let oldk = Kmanager.getK kman i in
	      let things = (Knowledge.getMPKeys oldk) in
	      let res = map things (function x -> (x, Assignment.add (Assignment.void()) (mm) x)) in
	      res
	  | SVar(v,i,PMKey) ->
	      let oldk = Kmanager.getK kman i in
	      let things = (Knowledge.getPMKeys oldk) in
	      let res = map things (function x -> (x, Assignment.add (Assignment.void()) (mm) x)) in
	      res

	  | Publ(SVar(v,i,RName)) ->
	      let oldk = Kmanager.getK kman i in
	      let things = (Knowledge.getRNames oldk) in
	      let res = map things (function PN(pn,pi) -> (PNp(pn,pi), Assignment.add (Assignment.void()) (SVar(v,i,RName)) (PN(pn,pi)) )) in
	      res
	  | Priv(SVar(v,i,RName)) ->
	      let oldk = Kmanager.getK kman i in
	      let things = (Knowledge.getRNames oldk) in
	      let res = map things (function PN(pn,pi) -> (PNm(pn,pi), Assignment.add (Assignment.void()) (SVar(v,i,RName)) (PN(pn,pi)) )) in
	      res

	  | Publ(SVar(v,i,Name)) ->
	      let oldk = Kmanager.getK kman i in
	      let things = (Knowledge.getAllNames oldk) in
	      let res = map things (function PN(pn,pi) -> (PNp(pn,pi), Assignment.add (Assignment.void()) (SVar(v,i,RName)) (PN(pn,pi)) )) in
	      res
	  | Priv(SVar(v,i,Name)) ->
	      let oldk = Kmanager.getK kman i in
	      let things = (Knowledge.getAllNames oldk) in
	      let res = map things (function PN(pn,pi) -> (PNm(pn,pi), Assignment.add (Assignment.void()) (SVar(v,i,RName)) (PN(pn,pi)) )) in
	      res

	  | Couple(m1, m2) ->
	      let res1 = mu_eq kman k asg m1 in
	      let res2 = (map (map res1 (function (a,b) ->  b)) (function sigma -> mu_eq kman k (asg) (sigma m2))) in
	      (rebuild res1 res2 (function (m1,s1) -> function (m2,s2) -> (Couple(m1,m2), s1*s2)))

	  | Crypt(m, K(s,i)) -> (
	      let res =
		if (Knowledge.derives k (asg) (K(s,i))) then map (mu_eq kman k asg m) (function (m1,s1) -> (Crypt(m1, K(s,i)), s1))
		else []
	      in
	      let kl = (Knowledge.getCrypts k (K(s,i))) in
	      res@(mulist_eq kman k asg kl m)
	     )
	  | Crypt(m, PNp(s,p)) ->  (
	      let res =
		if (Knowledge.derives k (asg) (PNm(s,p))) then map (mu_eq kman k asg m) (function (m1,s1) -> (Crypt(m1, PNp(s,p) ), s1))
		else []
	      in
	      let kl = (Knowledge.getCrypts k (PNm(s,p))) in
	      res@(mulist_eq kman k asg kl m)
	     )
	  | Crypt(m, PNm(s,p)) -> (
	      let res =
		if (Knowledge.derives k (asg) (PNp(s,p))) then map (mu_eq kman k asg m) (function (m1,s1) -> (Crypt(m1, PNm(s,p)), s1))
		else []
	      in
	      let kl = (Knowledge.getCrypts k (PNp(s,p))) in
	      res@(mulist_eq kman k asg kl m)
	     )

	  | Crypt(m, SVar(v,i,t)) ->
	      let res = mu_eq kman k asg (SVar(v,i,t)) in
	      let res2 = map res (function (a,b) -> mu_eq kman k asg (b m)) in
	      rebuild res res2 (function (a,b) -> function (c,d) -> (Crypt(c,a), b*d) )

	  | Crypt(m, Publ(SVar(v,i,t))) ->
	      let res = mu_eq kman k asg (Publ(SVar(v,i,t))) in
	      let res2 = map res (function (a,b) -> mu_eq kman k asg (b m)) in
	      rebuild res res2 (function (a,b) -> function (c,d) -> (Crypt(c,a), b*d) )

	  | Crypt(m, Priv(SVar(v,i,t))) ->
	      let res = mu_eq kman k asg (Priv(SVar(v,i,t))) in
	      let res2 = map res (function (a,b) -> mu_eq kman k asg (b m)) in
	      rebuild res res2 (function (a,b) -> function (c,d) -> (Crypt(c,a), b*d) )


	  | _ -> []
	 )
and mulist_eq =
  function kman ->
    function k ->
      function asg ->
	function cc ->
	  function n ->
	    match cc with
	    | Crypt(m,kk)::ccs ->
		let rr = ni_eq kman asg m n in
		(map rr (function r -> (Crypt(m,kk),r)))@(mulist_eq kman k asg ccs n)
	    | [] -> []
and ni_eq =
  function kman ->
    function asg ->
      function m ->
	function n ->
	  match m,n with
	  | K(_), _ when n=m ->  [(Assignment.void())]
	  | PN(_,_), _  when n=m ->  [(Assignment.void())]
	  | PNp(_,_), _ when n=m ->  [(Assignment.void())]
	  | PNm(_,_), _ when n=m ->  [(Assignment.void())]
	  | NO(_,_), _ when n=m ->  [(Assignment.void())]
	  | Crypt(mm,K(s,i)), Crypt(nn,K(w,y)) when (s=w && i=y) -> ni_eq kman asg mm nn
	  | Crypt(mm,PNp(s1,p1)), Crypt(nn,PNp(s2,p2)) when (s1=s2 && p1=p2) -> ni_eq kman asg mm nn
	  | Crypt(mm,PNm(s1,p1)), Crypt(nn,PNm(s2,p2)) when (s1=s2 && p1=p2) -> ni_eq kman asg mm nn
	  | Crypt(mm, SVar(v1,i,t1)), Crypt(nn, SVar(v2,j,t2)) when (t1=SKey || t1=MPKey || t1=PMKey)&&(t1=t2) ->
	      let res = ni_eq kman asg (SVar(v1,i,t1)) (SVar(v2,j,t2)) in
	      let res2 = map res (function sigma -> ni_eq kman asg (sigma mm) (sigma nn)) in
	      rebuild res res2 (function (s1) -> function (s2) -> (s1*s2))

	  | Couple(mm1,mm2), Couple(nn1,nn2) ->
	      let res1 = ni_eq kman asg mm1 nn1 in
	      let res2 = (map res1 (function sigma -> ni_eq kman (asg) (sigma mm2) (sigma nn2))) in
	      rebuild res1 res2 (function (s1) -> function (s2) -> (s1*s2))

	  | Var(v1), Var(v2) -> 
	      [Assignment.add (Assignment.void()) m n]
	  | Var(v1), _ -> 
	      [Assignment.add (Assignment.void()) m n]
	  | _, Var(v2) -> 
	      [Assignment.add (Assignment.void()) n m]

	  | SVar(v1,i,Gen), SVar(v2,j,Gen) ->
	      if n=m then [Assignment.void()]
	      else
		let min = if i<j then i else j in
		[Assignment.add (Assignment.add (Assignment.void()) m (SVar(v1,min,Gen))) n (SVar(v1,min,Gen))]
	  | SVar(v1,i,GenR), SVar(v2,j,GenR) ->
	      if n=m then [Assignment.void()]
	      else
		let min = if i<j then i else j in
		[Assignment.add (Assignment.add (Assignment.void()) m (SVar(v1,min,GenR))) n (SVar(v1,min,GenR))]
	  | SVar(v1,i,Name), SVar(v2,j,Name) ->
	      let min = if i<j then i else j in
	      let oldk = Kmanager.getK kman min in
	      let sk = Knowledge.getAllNames oldk in
	      let ss = map sk (function x -> Assignment.add (Assignment.add (Assignment.void()) (m) (x) ) (n) (x)) in
	      ss
	  | SVar(v1,i,RName), SVar(v2,j,RName) ->
	      let min = if i<j then i else j in
	      let oldk = Kmanager.getK kman min in
	      let sk = Knowledge.getRNames oldk in
	      let ss = map sk (function x -> Assignment.add (Assignment.add (Assignment.void()) (m) (x) ) (n) (x)) in
	      ss
	  | SVar(v1,i,SKey), SVar(v2,j,SKey) ->
	      let min = if i<j then i else j in
	      let oldk = Kmanager.getK kman min in
	      let sk = Knowledge.getSKeys oldk in
	      let ss = map sk (function x -> Assignment.add (Assignment.add (Assignment.void()) (m) (x) ) (n) (x)) in
	      ss
	  | SVar(v1,i,MPKey), SVar(v2,j,MPKey) ->
	      let min = if i<j then i else j in
	      let oldk = Kmanager.getK kman min in
	      let sk = Knowledge.getMPKeys oldk in
	      let ss = map sk (function x -> Assignment.add (Assignment.add (Assignment.void()) (m) (x) ) (n) (x)) in
	      ss
	  | SVar(v1,i,PMKey), SVar(v2,j,PMKey) ->
	      let min = if i<j then i else j in
	      let oldk = Kmanager.getK kman min in
	      let sk = Knowledge.getPMKeys oldk in
	      let ss = map sk (function x -> Assignment.add (Assignment.add (Assignment.void()) (m) (x) ) (n) (x)) in
	      ss

	  | SVar(v1,i,t1), SVar(v2,j,t2) when comparable(t1,t2) ->
	      let min = if i<j then i else j in
	      let oldk = (Kmanager.getK kman min) in
	      let mintg = mintype(t1,t2) in
	      let tasg = Assignment.add (Assignment.add (Assignment.void()) m (SVar(v1,min,mintg))) n (SVar(v1,min,mintg)) in
	      let mint = if min=i then t1 else t2 in
	      let minm = if mintg=t1 then SVar(v1,min,t1) else SVar(v2,min,t2) in
	      let maxm = if mintg=t1 then SVar(v2,min,t2) else SVar(v1,min,t1) in
	      let res = mu_eq kman oldk asg (minm) in
	      map res (function (mm,ss) -> tasg*ss*(Assignment.add (Assignment.add (Assignment.void()) maxm mm ) minm mm))

	  | Publ(SVar(v1,i,t1)), Publ(SVar(v2,j,t2)) when comparable(t1,t2) ->
	      let min = if i<j then i else j in
	      [Assignment.add (Assignment.add (Assignment.void()) (SVar(v1,i,t1)) (SVar(v1,min,mintype(t1,t2)))) (SVar(v2,j,t2)) (SVar(v1,min,mintype(t1,t2)))]
	  | Priv(SVar(v1,i,t1)), Priv(SVar(v2,j,t2)) when comparable(t1,t2) ->
	      let min = if i<j then i else j in
	      [Assignment.add (Assignment.add (Assignment.void()) (SVar(v1,i,t1)) (SVar(v1,min,mintype(t1,t2)))) (SVar(v2,j,t2)) (SVar(v1,min,mintype(t1,t2)))]


	  | SVar(v1,i,t1), Publ(SVar(v2,j,t2)) when (t1=Gen || t1=GenR)&&(t2=Name || t2=RName) ->
	      let min = if i<j then i else j in
	      [Assignment.add (Assignment.add (Assignment.void()) m (Publ(SVar(v1,min,t2)))) (SVar(v2,j,t2)) ((SVar(v2,min,t2)))]
	  | SVar(v1,i,t1), Priv(SVar(v2,j,t2)) when (t1=Gen || t1=GenR)&&(t2=Name || t2=RName) ->
	      let min = if i<j then i else j in
	      [Assignment.add (Assignment.add (Assignment.void()) m (Priv(SVar(v1,min,t2)))) (SVar(v2,j,t2)) ((SVar(v2,min,t2)))]
	  | Publ(SVar(v2,j,t2)), SVar(v1,i,t1)  when (t1=Gen || t1=GenR)&&(t2=Name || t2=RName) ->
	      let min = if i<j then i else j in
	      [Assignment.add (Assignment.add (Assignment.void()) (SVar(v2,j,t2)) ((SVar(v1,min,t2)))) n (Publ((SVar(v2,min,t2))))]
	  | Priv(SVar(v2,j,t2)), SVar(v1,i,t1)  when (t1=Gen || t1=GenR)&&(t2=Name || t2=RName) ->
	      let min = if i<j then i else j in
	      [Assignment.add (Assignment.add (Assignment.void()) (SVar(v2,j,t2)) ((SVar(v1,min,t2)))) n (Priv((SVar(v2,min,t2))))]

	  | sm, SVar(v,i,Gen)  when notSVar(sm) ->
	      let res = mu_eq kman (Kmanager.getK kman i) asg (m) in
	      map res (function (mm,ss) -> (ss*(Assignment.add (Assignment.void()) n mm)))
	  | sm, SVar(v,i,GenR)  when notSVar(sm) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu_eq kman oldk asg (m) in
	      let res2 = filter res (function (mm,ss) -> Knowledge.knows oldk  mm) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) n mm))
	  | sm, SVar(v,i,SKey)  when notSVar(sm) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu_eq kman oldk asg (m) in
	      let res2 = filter res (function (mm,ss) -> ((isskey mm))) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) n mm))
	  | sm, SVar(v,i,Name)  when notSVar(sm) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu_eq kman oldk asg (m) in
	      let res2 = filter res (function (mm,ss) -> ((isname mm))) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) n mm))
	  | sm, SVar(v,i,RName)  when notSVar(sm) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu_eq kman oldk asg (m) in
	      let res2 = filter res (function (mm,ss) -> (Knowledge.isRName oldk mm )) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) n mm))
	  | sm, SVar(v,i,t)  when (notSVar(sm) && (t=MPKey ||  t=PMKey)) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu_eq kman oldk asg (m) in
	      let res2 = filter res (function (mm,ss) -> Knowledge.isPMKey oldk mm) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) n mm))

	  | SVar(v,i,Gen), sm  when notSVar(sm) ->
	      let res = mu_eq kman (Kmanager.getK kman i) asg (n) in
	      map res (function (mm,ss) -> (ss*(Assignment.add (Assignment.void()) m mm)))
	  | SVar(v,i,GenR), sm  when notSVar(sm) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu_eq kman oldk asg (n) in
	      let res2 = filter res (function (mm,ss) -> Knowledge.knows oldk mm) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) m mm))
	  | SVar(v,i,SKey), sm  when notSVar(sm) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu_eq kman oldk asg (n) in
	      let res2 = filter res (function (mm,ss) -> ((isskey mm))) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) m mm))
	  | SVar(v,i,Name), sm  when notSVar(sm) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu_eq kman oldk asg (n) in
	      let res2 = filter res (function (mm,ss) -> ((isname mm))) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) m mm))
	  | SVar(v,i,RName), sm  when notSVar(sm) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu_eq kman oldk asg (n) in
	      let res2 = filter res (function (mm,ss) -> (Knowledge.isRName oldk mm )) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) m mm))
	  | SVar(v,i,t), sm  when (notSVar(sm) && (t=MPKey ||  t=PMKey)) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu_eq kman oldk asg (n) in
	      let res2 = filter res (function (mm,ss) -> Knowledge.isPMKey oldk mm) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) m mm))
	  | SVar(v,i,_), sm when notSVar(sm) ->
	      let res = mu_eq kman  (Kmanager.getK kman i) asg (n) in
	      map res (function (mm,ss) -> (ss*(Assignment.add (Assignment.void()) m mm)))
	  | _ -> []


(*fast unification functions*)

(* unification functions *)
let rec mu_eq =
  function kman ->
    function k ->
      function asg ->
	function mm -> (
	  match mm with
	  | Var(v) -> if (Knowledge.hasSKeys (Kmanager.getK kman  1)) then [( SVar(v,1,SKey),  (Assignment.add (Assignment.void()) mm (SVar(v,1,SKey)) ))] else []
	  | K(_) when (Knowledge.derives k (asg) mm) ->  [(mm, Assignment.void())]
	  | PN(_,_) when (Knowledge.derives k (asg) mm) ->  [(mm, Assignment.void())]
	  | PNp(_,_) when (Knowledge.derives k (asg) mm) ->  [(mm, Assignment.void())]
	  | PNm(_,_) when (Knowledge.derives k (asg) mm) ->  [(mm, Assignment.void())]
	  | NO(_,_) when (Knowledge.derives k (asg) mm) ->  [(mm, Assignment.void())]
	  | SVar(v,i,Gen) -> [(mm, Assignment.void())]
	  | SVar(v,i,GenR) -> [(mm, Assignment.void())]

	  | SVar(v,i,Name) ->
	      let oldk = Kmanager.getK kman i in
	      let things = Knowledge.getAllNames oldk in
	      let res = map things (function x -> (x, Assignment.add (Assignment.void()) (mm) x)) in
	      res
	  | SVar(v,i,RName) ->
	      let oldk = Kmanager.getK kman i in
	      let things = Knowledge.getRNames oldk in
	      let res = map things (function x -> (x, Assignment.add (Assignment.void()) (mm) x)) in
	      res
	  | SVar(v,i,SKey) ->
	      let oldk = Kmanager.getK kman i in
	      let things = Knowledge.getSKeys oldk in
	      let res = map things (function x -> (x, Assignment.add (Assignment.void()) (mm) x)) in
	      res
	  | SVar(v,i,MPKey) ->
	      let oldk = Kmanager.getK kman i in
	      let things = (Knowledge.getMPKeys oldk) in
	      let res = map things (function x -> (x, Assignment.add (Assignment.void()) (mm) x)) in
	      res
	  | SVar(v,i,PMKey) ->
	      let oldk = Kmanager.getK kman i in
	      let things = (Knowledge.getPMKeys oldk) in
	      let res = map things (function x -> (x, Assignment.add (Assignment.void()) (mm) x)) in
	      res

	  | Publ(SVar(v,i,RName)) ->
	      let oldk = Kmanager.getK kman i in
	      let things = (Knowledge.getRNames oldk) in
	      let res = map things (function PN(pn,pi) -> (PNp(pn,pi), Assignment.add (Assignment.void()) (SVar(v,i,RName)) (PN(pn,pi)) )) in
	      res
	  | Priv(SVar(v,i,RName)) ->
	      let oldk = Kmanager.getK kman i in
	      let things = (Knowledge.getRNames oldk) in
	      let res = map things (function PN(pn,pi) -> (PNm(pn,pi), Assignment.add (Assignment.void()) (SVar(v,i,RName)) (PN(pn,pi)) )) in
	      res

	  | Publ(SVar(v,i,Name)) ->
	      let oldk = Kmanager.getK kman i in
	      let things = (Knowledge.getAllNames oldk) in
	      let res = map things (function PN(pn,pi) -> (PNp(pn,pi), Assignment.add (Assignment.void()) (SVar(v,i,RName)) (PN(pn,pi)) )) in
	      res
	  | Priv(SVar(v,i,Name)) ->
	      let oldk = Kmanager.getK kman i in
	      let things = (Knowledge.getAllNames oldk) in
	      let res = map things (function PN(pn,pi) -> (PNm(pn,pi), Assignment.add (Assignment.void()) (SVar(v,i,RName)) (PN(pn,pi)) )) in
	      res

	  | Couple(m1, m2) ->
	      let res1 = mu_eq kman k asg m1 in
	      let res2 = (map (map res1 (function (a,b) ->  b)) (function sigma -> mu_eq kman k (asg) (sigma m2))) in
	      (rebuild res1 res2 (function (m1,s1) -> function (m2,s2) -> (Couple(m1,m2), s1*s2)))

	  | Crypt(m, K(s,i)) -> (
	      let res =
		if (Knowledge.derives k (asg) (K(s,i))) then map (mu_eq kman k asg m) (function (m1,s1) -> (Crypt(m1, K(s,i)), s1))
		else []
	      in
	      let kl = (Knowledge.getCrypts k (K(s,i))) in
	      res@(mulist_eq kman k asg kl m)
	     )
	  | Crypt(m, PNp(s,p)) ->  (
	      let res =
		if (Knowledge.derives k (asg) (PNm(s,p))) then map (mu_eq kman k asg m) (function (m1,s1) -> (Crypt(m1, PNp(s,p) ), s1))
		else []
	      in
	      let kl = (Knowledge.getCrypts k (PNm(s,p))) in
	      res@(mulist_eq kman k asg kl m)
	     )
	  | Crypt(m, PNm(s,p)) -> (
	      let res =
		if (Knowledge.derives k (asg) (PNp(s,p))) then map (mu_eq kman k asg m) (function (m1,s1) -> (Crypt(m1, PNm(s,p)), s1))
		else []
	      in
	      let kl = (Knowledge.getCrypts k (PNp(s,p))) in
	      res@(mulist_eq kman k asg kl m)
	     )

	  | Crypt(m, SVar(v,i,t)) ->
	      let res = mu_eq kman k asg (SVar(v,i,t)) in
	      let res2 = map res (function (a,b) -> mu_eq kman k asg (b m)) in
	      rebuild res res2 (function (a,b) -> function (c,d) -> (Crypt(c,a), b*d) )

	  | Crypt(m, Publ(SVar(v,i,t))) ->
	      let res = mu_eq kman k asg (Publ(SVar(v,i,t))) in
	      let res2 = map res (function (a,b) -> mu_eq kman k asg (b m)) in
	      rebuild res res2 (function (a,b) -> function (c,d) -> (Crypt(c,a), b*d) )

	  | Crypt(m, Priv(SVar(v,i,t))) ->
	      let res = mu_eq kman k asg (Priv(SVar(v,i,t))) in
	      let res2 = map res (function (a,b) -> mu_eq kman k asg (b m)) in
	      rebuild res res2 (function (a,b) -> function (c,d) -> (Crypt(c,a), b*d) )


	  | _ -> []
	 )
and mulist_eq =
  function kman ->
    function k ->
      function asg ->
	function cc ->
	  function n ->
	    match cc with
	    | Crypt(m,kk)::ccs ->
		let rr = ni_eq kman asg m n in
		(map rr (function r -> (Crypt(m,kk),r)))@(mulist_eq kman k asg ccs n)
	    | [] -> []
and ni_eq =
  function kman ->
    function asg ->
      function m ->
	function n ->
	  match m,n with
	  | K(_), _ when n=m ->  [(Assignment.void())]
	  | PN(_,_), _  when n=m ->  [(Assignment.void())]
	  | PNp(_,_), _ when n=m ->  [(Assignment.void())]
	  | PNm(_,_), _ when n=m ->  [(Assignment.void())]
	  | NO(_,_), _ when n=m ->  [(Assignment.void())]
	  | Crypt(mm,K(s,i)), Crypt(nn,K(w,y)) when (s=w && i=y) -> ni_eq kman asg mm nn
	  | Crypt(mm,PNp(s1,p1)), Crypt(nn,PNp(s2,p2)) when (s1=s2 && p1=p2) -> ni_eq kman asg mm nn
	  | Crypt(mm,PNm(s1,p1)), Crypt(nn,PNm(s2,p2)) when (s1=s2 && p1=p2) -> ni_eq kman asg mm nn
	  | Crypt(mm, SVar(v1,i,t1)), Crypt(nn, SVar(v2,j,t2)) when (t1=SKey || t1=MPKey || t1=PMKey)&&(t1=t2) ->
	      let res = ni_eq kman asg (SVar(v1,i,t1)) (SVar(v2,j,t2)) in
	      let res2 = map res (function sigma -> ni_eq kman asg (sigma mm) (sigma nn)) in
	      rebuild res res2 (function (s1) -> function (s2) -> (s1*s2))

	  | Couple(mm1,mm2), Couple(nn1,nn2) ->
	      let res1 = ni_eq kman asg mm1 nn1 in
	      let res2 = (map res1 (function sigma -> ni_eq kman (asg) (sigma mm2) (sigma nn2))) in
	      rebuild res1 res2 (function (s1) -> function (s2) -> (s1*s2))

	  | Var(v1), Var(v2) -> 
	      [Assignment.add (Assignment.void()) m n]
	  | Var(v1), _ -> 
	      [Assignment.add (Assignment.void()) m n]
	  | _, Var(v2) -> 
	      [Assignment.add (Assignment.void()) n m]

	  | SVar(v1,i,Gen), SVar(v2,j,Gen) ->
	      if n=m then [Assignment.void()]
	      else
		let min = if i<j then i else j in
		[Assignment.add (Assignment.add (Assignment.void()) m (SVar(v1,min,Gen))) n (SVar(v1,min,Gen))]
	  | SVar(v1,i,GenR), SVar(v2,j,GenR) ->
	      if n=m then [Assignment.void()]
	      else
		let min = if i<j then i else j in
		[Assignment.add (Assignment.add (Assignment.void()) m (SVar(v1,min,GenR))) n (SVar(v1,min,GenR))]
	  | SVar(v1,i,Name), SVar(v2,j,Name) ->
	      let min = if i<j then i else j in
	      let oldk = Kmanager.getK kman min in
	      let sk = Knowledge.getAllNames oldk in
	      let ss = map sk (function x -> Assignment.add (Assignment.add (Assignment.void()) (m) (x) ) (n) (x)) in
(	      (*print_list ss (Assignment.prnt);*)
	      ss)
	  | SVar(v1,i,RName), SVar(v2,j,RName) ->
	      let min = if i<j then i else j in
	      let oldk = Kmanager.getK kman min in
	      let sk = Knowledge.getRNames oldk in
	      let ss = map sk (function x -> Assignment.add (Assignment.add (Assignment.void()) (m) (x) ) (n) (x)) in
	      ss
	  | SVar(v1,i,SKey), SVar(v2,j,SKey) ->
	      let min = if i<j then i else j in
	      let oldk = Kmanager.getK kman min in
	      let sk = Knowledge.getSKeys oldk in
	      let ss = map sk (function x -> Assignment.add (Assignment.add (Assignment.void()) (m) (x) ) (n) (x)) in
	      ss
	  | SVar(v1,i,MPKey), SVar(v2,j,MPKey) ->
	      let min = if i<j then i else j in
	      let oldk = Kmanager.getK kman min in
	      let sk = Knowledge.getMPKeys oldk in
	      let ss = map sk (function x -> Assignment.add (Assignment.add (Assignment.void()) (m) (x) ) (n) (x)) in
	      ss
	  | SVar(v1,i,PMKey), SVar(v2,j,PMKey) ->
	      let min = if i<j then i else j in
	      let oldk = Kmanager.getK kman min in
	      let sk = Knowledge.getPMKeys oldk in
	      let ss = map sk (function x -> Assignment.add (Assignment.add (Assignment.void()) (m) (x) ) (n) (x)) in
	      ss

	  | SVar(v1,i,t1), SVar(v2,j,t2) when comparable(t1,t2) ->
	      let min = if i<j then i else j in
	      let oldk = (Kmanager.getK kman min) in
	      let mintg = mintype(t1,t2) in
	      let tasg = Assignment.add (Assignment.add (Assignment.void()) m (SVar(v1,min,mintg))) n (SVar(v1,min,mintg)) in
	      let mint = if min=i then t1 else t2 in
	      let minm = if mintg=t1 then SVar(v1,min,t1) else SVar(v2,min,t2) in
	      let maxm = if mintg=t1 then SVar(v2,min,t2) else SVar(v1,min,t1) in
	      let res = mu_eq kman oldk asg (minm) in
	      map res (function (mm,ss) -> tasg*ss*(Assignment.add (Assignment.add (Assignment.void()) maxm mm ) minm mm))

	  | Publ(SVar(v1,i,t1)), Publ(SVar(v2,j,t2)) when comparable(t1,t2) ->
	      let min = if i<j then i else j in
	      [Assignment.add (Assignment.add (Assignment.void()) (SVar(v1,i,t1)) (SVar(v1,min,mintype(t1,t2)))) (SVar(v2,j,t2)) (SVar(v1,min,mintype(t1,t2)))]
	  | Priv(SVar(v1,i,t1)), Priv(SVar(v2,j,t2)) when comparable(t1,t2) ->
	      let min = if i<j then i else j in
	      [Assignment.add (Assignment.add (Assignment.void()) (SVar(v1,i,t1)) (SVar(v1,min,mintype(t1,t2)))) (SVar(v2,j,t2)) (SVar(v1,min,mintype(t1,t2)))]


	  | SVar(v1,i,t1), Publ(SVar(v2,j,t2)) when (t1=Gen || t1=GenR)&&(t2=Name || t2=RName) ->
	      let min = if i<j then i else j in
	      [Assignment.add (Assignment.add (Assignment.void()) m (Publ(SVar(v1,min,t2)))) (SVar(v2,j,t2)) ((SVar(v2,min,t2)))]
	  | SVar(v1,i,t1), Priv(SVar(v2,j,t2)) when (t1=Gen || t1=GenR)&&(t2=Name || t2=RName) ->
	      let min = if i<j then i else j in
	      [Assignment.add (Assignment.add (Assignment.void()) m (Priv(SVar(v1,min,t2)))) (SVar(v2,j,t2)) ((SVar(v2,min,t2)))]
	  | Publ(SVar(v2,j,t2)), SVar(v1,i,t1)  when (t1=Gen || t1=GenR)&&(t2=Name || t2=RName) ->
	      let min = if i<j then i else j in
	      [Assignment.add (Assignment.add (Assignment.void()) (SVar(v2,j,t2)) ((SVar(v1,min,t2)))) n (Publ((SVar(v2,min,t2))))]
	  | Priv(SVar(v2,j,t2)), SVar(v1,i,t1)  when (t1=Gen || t1=GenR)&&(t2=Name || t2=RName) ->
	      let min = if i<j then i else j in
	      [Assignment.add (Assignment.add (Assignment.void()) (SVar(v2,j,t2)) ((SVar(v1,min,t2)))) n (Priv((SVar(v2,min,t2))))]

	  | sm, SVar(v,i,Gen)  when notSVar(sm) ->
	      let res = mu_eq kman (Kmanager.getK kman i) asg (m) in
	      map res (function (mm,ss) -> (ss*(Assignment.add (Assignment.void()) n mm)))
	  | sm, SVar(v,i,GenR)  when notSVar(sm) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu_eq kman oldk asg (m) in
	      let res2 = filter res (function (mm,ss) -> Knowledge.knows oldk  mm) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) n mm))
	  | sm, SVar(v,i,SKey)  when notSVar(sm) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu_eq kman oldk asg (m) in
	      let res2 = filter res (function (mm,ss) -> ((isskey mm))) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) n mm))
	  | sm, SVar(v,i,Name)  when notSVar(sm) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu_eq kman oldk asg (m) in
	      let res2 = filter res (function (mm,ss) -> ((isname mm))) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) n mm))
	  | sm, SVar(v,i,RName)  when notSVar(sm) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu_eq kman oldk asg (m) in
	      let res2 = filter res (function (mm,ss) -> (Knowledge.isRName oldk mm )) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) n mm))
	  | sm, SVar(v,i,t)  when (notSVar(sm) && (t=MPKey ||  t=PMKey)) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu_eq kman oldk asg (m) in
	      let res2 = filter res (function (mm,ss) -> Knowledge.isPMKey oldk mm) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) n mm))

	  | SVar(v,i,Gen), sm  when notSVar(sm) ->
	      let res = mu_eq kman (Kmanager.getK kman i) asg (n) in
	      map res (function (mm,ss) -> (ss*(Assignment.add (Assignment.void()) m mm)))
	  | SVar(v,i,GenR), sm  when notSVar(sm) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu_eq kman oldk asg (n) in
	      let res2 = filter res (function (mm,ss) -> Knowledge.knows oldk mm) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) m mm))
	  | SVar(v,i,SKey), sm  when notSVar(sm) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu_eq kman oldk asg (n) in
	      let res2 = filter res (function (mm,ss) -> ((isskey mm))) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) m mm))
	  | SVar(v,i,Name), sm  when notSVar(sm) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu_eq kman oldk asg (n) in
	      let res2 = filter res (function (mm,ss) -> ((isname mm))) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) m mm))
	  | SVar(v,i,RName), sm  when notSVar(sm) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu_eq kman oldk asg (n) in
	      let res2 = filter res (function (mm,ss) -> (Knowledge.isRName oldk mm )) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) m mm))
	  | SVar(v,i,t), sm  when (notSVar(sm) && (t=MPKey ||  t=PMKey)) ->
	      let oldk = (Kmanager.getK kman i) in
	      let res = mu_eq kman oldk asg (n) in
	      let res2 = filter res (function (mm,ss) -> Knowledge.isPMKey oldk mm) in
	      map res2 (function (mm,ss) -> (ss*Assignment.add (Assignment.void()) m mm))
	  | SVar(v,i,_), sm when notSVar(sm) ->
	      let res = mu_eq kman  (Kmanager.getK kman i) asg (n) in
	      map res (function (mm,ss) -> (ss*(Assignment.add (Assignment.void()) m mm)))
	  | _ -> []

