(*
   KHash.mli
   29/08/03
   Giacomo Baldi
   
   Definition of Khash interface

*)

(**
   This module implements a hash table of weak pointers over kb identification numbers.
   A weak pointer is a reference to a datum which is not taken in account by the garbage collector
   when counting references. A weak pointer can be a reference to an object or to nothing if the object it was pointing to
   is already expired (removed by gc). This structure is useful in sharing kbs occuping only the least amount of memory.  
*)

open Knowledge

(**
   implemented as an array of arrays of weak pointers.
*)
type khash


(**
   init takes a int n.
   @return an empty khash table
*)
val init: int -> khash

(**
   clear takes a khash and a in sp.
   Starting from sp it clears all khash entries
*)
val clear: khash -> int -> unit

(**
   insert takes a khash and a kb and updates the table according to kb.
*)
val insert: khash -> kb -> unit

(**
   get takes a khash and a kb identification.
   @returns the corresponding kb if it is in the table, or a void one otherwise. 
*)
val get: khash -> int -> kb

(**
   size takes a khash
   @returns the maximum size of a bucket in the table.
*)
val size: khash -> int
