(*
	kmanager.mli
	05/04/2003
	Giacomo Baldi

	Definition of a knowledge manager.
*)

(**
   This module handle the knowledge pool. Creation and expansion of knowledge are permitted only 
   using kmanager functions. 
*)

open Message
open Knowledge
open Assignment


(**
   When searching for a alredy expired knowledge.
*)
exception NoKnowledge

(**
   implemented using Khash module
*)
type km

(**
   initKM takes an int n.
   @return a km of size n. If n is too big, the biggest possible km is created (Sys.max_array_length)
*)
val initKM: int -> km

(**
   getK takes a km and a kb identification number k.
   @return the correct kb.
   @raise NoKnowledge if a kb with k as identification is not in km
*)
val getK: km -> int -> kb

(**
   newK takes a km.
   @return a new kb with a fresh identification number.
*)
val newK: km -> kb

(**
   expandK takes a km, a kb, an assignment and a message m.
   @return a kb expanding the given one (under assignment) with message m.
*)
val expandK: km*kb*assignment*msg -> kb


(**
   updateK takes a km, a kb and an assignment a.
   @return an new kb update with a.
*)
val updateK: km*kb*assignment -> kb

(**
   size takes a km.
   @return statistic about km data structure.
*)
val size: km -> int

