open Message;;
open Sys;;
open Utils;;
open Node
open Parsecmdline;;
open Gc;;
open Prenex;;
open Evalh;;
open Misc;;
open Tracer;;

let opts = init_options();;
let continue=ref(true);;

if not (!Sys.interactive) then
  (
   if Array.length(Sys.argv)<2 then (synopsis false; continue:=false)
   else (
     (parsecmdline (Sys.argv) opts);
     if (opts.errmsg<>"") then (print_string("\n"^opts.errmsg^"\n"); continue:=false)
     else
       if opts.help then (synopsis true; continue:=false)
       else
	 if opts.pr_file="" then (synopsis false; continue:=false)
	 else (if opts.pl_file="" then (synopsis false; continue:=false) else (if opts.kb_file="" then (synopsis false; continue:=false) else ()))
    )
  )
else (synopsis false; continue:=false);;



let rec readall =
  function ct ->
    if (Parser.more_parsing()) then
      let p = Parser.parsePrincipal() in
      if (Parser.getErrorCode()<>0) then (print_string(Parser.getErrorMsg()); failwith "\nSyntax error in parsing principals\n")
      else ( let ct2 = p::ct in readall(ct2))
    else ct;;

if(!continue) then (
  
  (* Parsing *)


  Pl_parser.init_parser(opts.pl_file);
  let pl = Pl_parser.parseFormula() in 
  if (Pl_parser.getErrorCode()<>0) then (print_string(Pl_parser.getErrorMsg()); failwith "\nSyntax error in parsing formula\n")
  else ();
  Pl_parser.done_parser();
  

  Pl_parser.init_parser(opts.pj_file);
  let plj = Pl_parser.parseFormula() in 
  if (Pl_parser.getErrorCode()<>0) then (print_string(Pl_parser.getErrorMsg()); failwith "\nSyntax error in parsing join components\n")
  else ();

(*Logic.prnt plj;*)

  Pl_parser.done_parser();

  Pl_parser.init_parser(opts.kb_file);
  let kbl = Pl_parser.parseKb() in 
  if (Pl_parser.getErrorCode()<>0) then (print_string(Pl_parser.getErrorMsg()); failwith "\nSyntax error in parsing knowledge\n")
  else ();
  Pl_parser.done_parser();

  Parser.init_parser(opts.pr_file);
  let prns = readall([]) in
  Parser.done_parser();

(* Searching *)


  let prn = extract prns in 
  let roles = List.length prn in 
  let mytree = gen_tree (opts.max_princ) roles prn in
  let wetree= ntree mytree pl in 
  let weitree = trace_value wetree (Int 0) in weitree;
  let slist = sort_trace weitree in
  let nlist = sort_ccc slist in nlist;
  let clist = max_perm nlist in 
  let mylist = rem_per clist in 
  let ss = init_search prns kbl opts in  
  do_joins ss pl plj mylist;

  if (opts.output_type = 1) then Gviz.initgraph "prova" else ();
  if (opts.output_type = 2 && opts.attack_type = 2) then Parsecmdline.htmlinit() else ();
  do_search ss;
  if (opts.output_type = 1) then Gviz.donegraph() else ();
  if (opts.output_type = 2 && opts.attack_type = 2) then Parsecmdline.htmldone() else ();

  (*if (opts.feedback) then *)prerr_stats ss;(* else ()*)
    


(****prova partition****)
(*
let l = 0::1::2::3::[] in
let pp = partition l in
print_list pp (function ll -> 
  print_string("---\n"); print_list ll (function lll -> 
    print_list lll (function x -> print_string(string_of_int(x))); print_string("*"));
      print_string("\n"))

*)
 ) else ()






