(*
   khash.ml
   27/09/03
   Giacomo Baldi

   Implementation of a weak hash table of kb

*)

open Knowledge

type khash = (kb Weak.t) array

let rec clear =
  function hsh ->
    function n ->
      if n= Array.length hsh then ()
      else (Array.set hsh n (Weak.create 1); clear hsh (n+1)) 

let hash =
  function key ->
    function hsh ->
      let (f,i) = modf(0.6180339887 *. float_of_int(key)) in
      int_of_float( f *. float_of_int(Array.length hsh))

let init =
  function size -> 
    let bkt = Weak.create 1 in
    let hsh = Array.make (min size (Sys.max_array_length-1)) bkt in
    (clear hsh 0; hsh)


let rec insertbkt =
  function bkt ->
    function kb ->
      function n ->
	if n = Weak.length bkt then (
	  let nbkt = Weak.create (n+1) in
	  Weak.set nbkt n (Some(kb));
	  Weak.blit bkt 0 nbkt 0 n;
	  nbkt
	 )
	else (
	  if (Weak.get bkt n = None) then (Weak.set bkt n (Some(kb)); bkt)
	  else insertbkt bkt kb (n+1)
	 )
	

let rec getbkt =
  function bkt ->
    function key ->
      function n ->
	if n = Weak.length bkt then ( Knowledge.void())
	else (
	  match Weak.get bkt n with
	  | Some(kb) -> if (Knowledge.id_of kb = key) then kb else getbkt bkt key (n+1)
	  | None -> getbkt bkt key (n+1)
	 )
      

let insert =
  function hsh ->
    function kb ->
      let key = Knowledge.id_of kb in
      let h = hash key hsh in
      let bkt = Array.get hsh h in
      let nbkt = insertbkt bkt kb 0 in
      Array.set hsh h nbkt

let get =
  function hsh ->
    function key ->
      let h = hash key hsh in
      let bkt = Array.get hsh h in
      getbkt bkt key 0
      

let size =
  function hsh ->
    Array.fold_right (function bkt -> function mm -> max (Weak.length bkt) mm) hsh 0

  
