(*
	Step.ml
	30/01/2003
	Giacomo Baldi

	implementation of module Step. It implements status transitions.
*)


open Node
open Context
open Status
open Process
open Principal
open Message
open Assignment
open Utils
open Action
open Core


let findPrincipalRedexes =
  function kman ->
    function k ->
      function asg ->
	function e ->
	  function act ->
	    function chs ->
	      let rec expand =
		function rds ->
		  function ex -> (
		    match rds with
		    | [] -> []
		    | (a,(nm, ii, vlist, expr))::xs when expr=Nil -> (a,(nm, ii, vlist, ex))::(expand xs ex)
		    | (a,(nm, ii, vlist, expr))::xs  -> (a,(nm, ii, vlist, Par(expr,ex)))::(expand xs ex)
		   )
	      in
	      let rec scan =
		function exp ->
		  match exp with
		  | Par(e1,e2) ->
		      let re1 = scan e1 in
		      let re2 = scan e2 in
		      (expand re1 e2) @ (expand re2 e1)
		  | Sum(e1,e2) ->
		      let re1 = scan e1 in
		      let re2 = scan e2  in
		      re1@re2
		  | Nil -> []
		  | ee when chs(ee) -> (act ee e kman k asg)
		  | _ -> []
	      in
	      let (nm,ii,vlist,expr) = e in scan expr



let lazyaction =
  function e ->
    function p ->
      function kman ->
	function k ->
	  function asg ->
	    let ll = l kman asg in
	    let mm = mu kman k in
	    let (nm,ii,vlist,expr) = p in
	    match e with
	    | In(m,e1) -> (*print_string("Checking input: "); prntMsg m; print_string("\n");*)
		(
		 let jjj = [] (*llazy k asg m m*) in

(*
   print_string("llazy: ");
   print_list jjj (Assignment.prnt);
 *)
		 let jjf = filter jjj (function s -> try (Principal.check (Principal.substitute s p)) with _ -> false) in
(*
   print_string("llazy checked: ");
   print_list jjf (Assignment.prnt);
 *)
		 let zzz = if jjf=[] then [Assignment.void()] else jjf in

		 let rrr = debox (map zzz (function s -> let ls = ll (s m) in mul [s] ls (function a -> function b -> a * b))) in
(*
   print_string("l-fy: ");
   print_list rrr (Assignment.prnt);
 *)
		 let nss = if rrr = [] then [Assignment.void()] else rrr in
		 let res = (map nss (function sigma -> mm (sigma) (sigma m))) in
(*
   print_string("m-fy: ");
   print_list res (function tl -> print_list tl (function (m1,s1) -> Assignment.prnt s1));
 *)
		 let mlm = rebuild nss res (function sigma -> function (m1,a) ->(Input(((nm,ii),m1,sigma*a)),(nm, ii, vlist, e1))) in
		 remclone mlm (function (x,p1) -> function (mes,p2) -> not (Action.equal x mes))
		)

	    | Out(m,e1) -> (*print_string("Checking output: "); prntMsg m; print_string("\n");*)
		(
		 let jjj = [] (*llazy k asg m m*) in
		 let jjf = filter jjj (function s -> try (Principal.check (Principal.substitute s p)) with _ -> false) in
		 let rrr = if jjf=[] then [Assignment.void()] else jjf in
		 let res = debox (map rrr (function s -> let ls = ll (s m) in mul [s] ls (function a -> function b -> a * b))) in
		 let trr = map res (function sigma -> (Output(((nm,ii),m, sigma)),(nm, ii, vlist, e1))) in
		 remclone trr (function (x,p1) -> function (mes,p2) -> not (Action.equal x mes))
		)

let isOUT =
  function x ->
    match x with
    | Out(_) -> true
    | _ -> false
let isIN =
  function x ->
    match x with
    | In(_) -> true
    | _ -> false


let getRedex =
  function node ->
    let np = Node.getnd node in
    Node.getRedex node


let generatelazyRedexesOut =
  function kman ->
    function k ->
      function asg ->
	function ct ->
	  function np ->
	    let op = (Context.getprincipal (ct, np)) in
	    let pp = Principal.substitute asg op in
	    map (((findPrincipalRedexes kman k (asg) pp lazyaction isOUT))) (function (b,c) -> (op,b,c))

let generatelazyRedexesIn =
  function kman ->
    function k ->
      function asg ->
	function ct ->
	  function np ->
	    let op = (Context.getprincipal (ct, np)) in
	    let pp = Principal.substitute asg op in
	    map (((findPrincipalRedexes kman k (asg) pp lazyaction isIN))) (function (b,c) -> (op,b,c))

(*
   Given a node we extract information in it. We try to find the ndth redex. If it is a
   - NoMoreAct, we set the node as "burned", all possible transitions have been made.
   - Input action, we get the message, make the correct assignment and create a new node with an updated
   status,  an updated assignment and the old knowledge
   - Output action, we fire the message and create a new node with an updated status, knowledge  and old assignment
   If we ge a NulNode as a result we know we have to try a join to have join transitions.
 *)

let rec transition =
  function frout ->
    function frin ->
      function kman ->
	function node ->
	  let (oldct, oldasg, oldkb) = Node.getstatus node in
	  let nd = Node.getnd node in
	  if nd = -1 then Node.nulNode()
	  else
	    (
	     let red = getRedex node in
	     match red  with
	     | (_, Closed, _) -> (Node.setnd (node,-1); Node.nulNode())
	     | (_, NoMoreAct, _) -> (
		 if (nd<(Context.howmany oldct)&&(Node.isOut node)) then (
		   Node.setnd (node,nd+1);
		   let outs = frout kman oldkb oldasg oldct (nd+1) in
		   if outs = [] then (
		     if (nd+1 = (Context.howmany oldct)) then (
		       Node.setnd(node,0);
		       Node.setIn node
		      ) else ();
		     transition frout frin  kman node
		    )
		   else (
		     Node.close node;
		     Node.settable (node, (outs));
		     transition frout frin kman node
		    )
		  )
		 else
		   if (nd<(Context.howmany oldct)&&( not (Node.isOut node))) then (
		     Node.setnd (node,nd+1);
		     let ins = frin kman oldkb oldasg oldct (nd+1) in
		     (
		      Node.settable (node, (ins));
		      transition  frout frin  kman node
		     )
		    )
		   else (
		     Node.setnd (node,-1);
		     Node.nulNode()
		    )
		)
	     | (p1, Output(ppn,m,sigma), p2) -> (
		 let newasg = oldasg * sigma in
		 let newct = Context.applyall (oldct ++ (p1, (Principal.substitute newasg p2))) newasg in
		 let newkb = (Kmanager.expandK(kman, oldkb, newasg, sigma m))  in
		 let newst = Status.create(newct, newasg, newkb) in
		 let newnode = Node.create(newst) in (
		 Node.setfather (newnode, node);
		 Node.setlabel (newnode, Output(ppn,m,sigma));
		 newnode)
		)
	     | (p1, Input(ppn,m,sigma), p2) -> (
		 let newasg = oldasg * sigma in
		 let newct = Context.applyall (oldct ++ (p1, (Principal.substitute newasg p2))) newasg in
		 let newkb = Kmanager.updateK(kman,oldkb,newasg) in
		 let newst = Status.create(newct, newasg, newkb) in
		 let newnode = Node.create(newst) in (
		 Node.setfather (newnode, node);
		 Node.setlabel (newnode, Input(ppn,m,sigma));
		 newnode)
		)
	    )

let getTransitionFunction =
  function km ->
     transition generatelazyRedexesOut generatelazyRedexesIn km
