(* Tokenizer *)


(* Definizione di identificatori e simboli*)
type ltok =
	| LIde of string
	| LSym of string
	| LEnd
	| LStop
	| LUnknown of char;;

(* Buffer di analisi della stringa *)
type lbuf = string*(int ref)*int;;

let forward buf = match buf with
	| (s,n,sz) -> n:= !n+1;;

let nextIs = fun (ch, (s,n,sz)) ->
	if !n=sz-1 then false
	else ch=s.[!n+1];;


let getIde buf = let (s,n,sz) = buf in
	let rec cntIde l = if l=sz then l
		else(
			match s.[l] with
				| 'A'..'Z'
				| 'a'..'z'
				| '0'..'9'
				| '_' -> cntIde (l+1)
				| _  -> l
			)
	in let k = cntIde !n
		in let res = String.sub s !n (k - !n)
			in (n:=k; res);;

let rec gettok buf =
	let analyze ch = (
		match ch with
			| ' '
			| '\t' | '\n' -> (forward buf; gettok buf)
			| 'A'..'Z'
			| 'a'..'z'
			| '0'..'9' -> LIde(getIde buf)
			| '(' | ')' | '.' | '|' | '+'
			| '{' | '}' | ',' | '?' | ';'
			| '-' | ':' | '[' | ']' | '&'
			| '|' | '>' | '=' | '<' | '!'
			| '/' | '*' -> (forward buf; LSym(String.make 1 ch))
			| _ -> (forward buf; LUnknown(ch))
		)
	in match buf with
		| (s,n,sz)  when !n<sz-> analyze s.[!n]
		| (s,n,sz) -> LEnd;;

let prnt h = match h with
	| LIde(s) -> print_string ("-->"^s^"\n")
	| LSym(s) -> print_string ("-->"^s^"\n")
	| LEnd -> print_string ("fine\n")
	| LStop -> print_string ("stop\n")
	| LUnknown(c) -> print_string ("Unknown:"^(String.make 1 c)^"\n");;


let mainbuf = ref(("",ref(0),0));;
let chan = ref(stdin);;
let cline = ref(0);;
let init_tokenizer fname = chan:= (open_in fname);;
let done_tokenizer = fun () -> close_in !chan;;


let flushline = fun () ->
	let nline = try (input_line !chan) with End_of_file -> "\n" in
		if (nline="\n") then (cline:=!cline+1; mainbuf:= (nline, ref(0), -1))
		else (cline:=!cline+1; mainbuf:= (nline, ref(0), String.length nline));;

let rec getToken = function () ->
	let res = gettok !mainbuf
		in match res with
			| LEnd -> (
				flushline();
				let (nline,n,sz) = !mainbuf in
					if sz<> -1 then getToken()
					else LStop
				)
			| _ -> res;;

let pushback s = match s with
	| LIde(k)
	| LSym(k) -> let (ss, p, nn) = !mainbuf in mainbuf:=(ss,ref(!p-String.length k), nn)
	| _ -> ();;

let where = fun () -> let (nline,n,sz) = !mainbuf in (!cline,!n);;


(* ---------------------------TRASH-------------------------------- *)

(*
init_tokenizer("p");;

let flag = ref(true);;
while (!flag) do
	let s1 = getToken() in
	if s1=LStop then flag:=false
	else prnt s1;
done;;
*)
(*pushback(LSym("."));;
let s1 = getToken();;
print_string "dddd\n";;
prnt s1;;
*)
(*
gettoken
			| LIde(s) -> print_string (s^"\n"); res
			| LSym(s) -> print_string (s^"\n"); res
			| LLCom -> print_string ("Comment\n"); res
			| LGCom(b) -> print_string ("Comment G\n"); res
			| LUnknown(c) -> print_string ("Unknown:"^(String.make 1 c)^"\n"); res
			| _ -> print_string "_\n"; res;;
*)
