(*
	Assignment.ml
	23/01/2003
	Giacomo Baldi

	Implementation of operations on assignments (Chi)
*)

open Message
open Utils


exception NoMeaning
exception NeedVariable

type assignment = msg -> msg

(************table version*************)


let getfromtable =
  function table ->
    function v ->
      let res = filter table (function (vv,mm) -> v=vv) in
      if res = [] then v else snd(head res)

let addtable =
  function table ->
    function (v,m) ->
      let rec closure =
	function tt ->
	  function (vv,mm) ->
	    let nt = filter tt (function (vvv,mmm) -> mmm=vv) in
	    let ad = map nt (function (vvv,mmm) -> (vvv,mm)) in
	    ad@(debox(map ad (function (vvv,mmm) -> closure tt (vvv,mmm))))
      in
      if (v=m) then table
      else
	match v with
	| Var(_)
	| SVar(_) -> (
	    let nm = getfromtable table v in
	    if nm = v then (v,m)::table
	    else table
	   )
	| _ -> raise NeedVariable




let rec f =
  function tt ->
    function n ->
      let rec pass =
	function table ->
	  function m ->
	    match m with
	    | Var(p) -> getfromtable table m
	    | SVar(p,i,t) -> getfromtable table m
	    | BVar(p) ->
		let nm =getfromtable table (Var(p)) in
		if nm = Var(p) then m else nm
	    | Couple(m1,m2) -> Couple(pass table m1, pass table m2)
	    | Crypt(m1,key) -> Crypt(pass table m1, pass table key)
	    | K(_)
	    | PN(_)
	    | PNp(_)
	    | PNm(_)
	    | NO(_) -> m
	    | Publ(Var(v)) ->
		let vv = getfromtable table (Var(v)) in (
		match vv with
		| PN(s,p) -> PNp(s,p)
		| SVar(vr,i,t) -> Publ(SVar(vr,i,t))
		| Var(v) -> m
		| _ ->  Publ(vv)
	       )
	    | Priv(Var(v)) ->
		let vv = getfromtable table (Var(v)) in (
		match vv with
		| PN(s,p) -> PNm(s,p)
		| SVar(vr,i,t) -> Priv(SVar(vr,i,t))
		| Var(v) -> m
		| _ ->  Priv(vv)
	       )
	    | Publ(SVar(v,i,t)) when (t=Gen || t=GenR)->
		let vv2 = getfromtable table (SVar(v,i,t)) in (
		match vv2 with
		| PN(s,p) -> PNp(s,p)
		| SVar(v,i,Name) -> Publ(vv2)
		| SVar(v,i,RName) -> Publ(vv2)
		| SVar(v,i,t) -> m
		| _ ->  Publ(vv2)
	       )
	    | Priv(SVar(v,i,t)) when (t=Gen || t=GenR)->
		let vv2 = getfromtable table (SVar(v,i,t)) in (
		match vv2 with
		| PN(s,p) -> PNm(s,p)
		| SVar(v,i,Name) -> Priv(vv2)
		| SVar(v,i,RName) -> Priv(vv2)
		| SVar(v,i,t) -> m
		| _ -> Priv(vv2)
	       )
	    | Publ(SVar(v,i,Name)) ->
		let vv2 = getfromtable table (SVar(v,i,Name)) in (
		match vv2 with
		| PN(s,p) -> PNp(s,p)
		| SVar(v,i,Name) -> m
		| SVar(v,i,RName) -> Publ(vv2)
		| _ -> Publ(vv2)
	       )
	    | Priv(SVar(v,i,Name)) ->
		let vv2 = getfromtable table (SVar(v,i,Name)) in (
		match vv2 with
		| PN(s,p) -> PNm(s,p)
		| SVar(v,i,Name) -> m
		| SVar(v,i,RName) -> Priv(vv2)
		| _ -> Priv(vv2)
	       )
	    | Publ(SVar(v,i,RName)) ->
		let vv2 = getfromtable table (SVar(v,i,RName)) in (
		match vv2 with
		| PN(s,p) -> PNp(s,p)
		| SVar(v,i,RName) -> m
		| _ -> Publ(vv2)
	       )
	    | Priv(SVar(v,i,RName)) ->
		let vv2 = getfromtable table (SVar(v,i,RName)) in (
		match vv2 with
		| PN(s,p) -> PNm(s,p)
		| SVar(v,i,RName) -> m
		| _ -> Priv(vv2)
	       )
	    | Priv(_)
	    | Publ(_) -> m
	    | Bottom -> m
	    | _ -> raise NoMeaning
      in
      match n with
      | Trick(_) -> Trick(tt)
      | _ -> 
	  let newm = pass tt n in
	  if newm = n then newm
	  else f tt newm



let void =
  function () ->
    f []

let newasg =
  function table ->
    let checktable = filter table (function (v,m) -> v<>m) in
    f checktable

let isVoid =
  function asg ->
    let Trick(tab) = asg (Trick([])) in
    tab = []

let add =
  function asg ->
    function var ->
      function newmsg ->
	let Trick(tab) = asg (Trick([])) in
	let newtable = addtable tab (var,newmsg) in
	f newtable


let prntHtml =
  function asg ->
    let rec prnttab =
      function l ->
	match l with
	| (v,n)::xs -> (prntMsgHtml ((v)); print_string(" -&gt "); prntMsgHtml n; print_string("<br>")); prnttab xs
	| [] -> print_string(" .<br>")
    in
    let Trick(tab) = asg (Trick([])) in
    prnttab tab


let prnt =
  function asg ->
    let rec prnttab =
      function l ->
	match l with
	| (v,n)::xs -> (prntMsg ((v)); print_string(" -> "); prntMsg n; print_string("\n")); prnttab xs
	| [] -> print_string(" .\n")
    in
    let Trick(tab) = asg (Trick([])) in
    prnttab tab


let tostring =
  function asg ->
    let rec prnttab =
      function ll ->
	match ll with
	| [] -> ("\\n");
	| (v,n)::xs -> (Message.tostring v)^" -> "^(Message.tostring n)^("\\n")^(prnttab xs)
    in
    let Trick(tab) = asg (Trick([])) in
    prnttab tab


let ( * ) =
  function alfa ->
    function beta ->
      function m ->
	let Trick(t1) = alfa (Trick([])) in
	let Trick(t2) = beta (Trick([])) in
	let nt = t1@t2 in
	newasg nt m

let comp = ( * )

let listfy = 
  function asg ->
    let Trick(tab) = asg (Trick([])) in
    tab

let leftof =
  function asg ->
    function v ->
      let sss = listfy asg in
      map (filter sss (function (a,b) -> b=v)) (fst)
	
