(*
	Context.ml
	30/01/2003
	Giacomo Baldi
	
	Implementation of a context
*)

(***** Optimize structure!!!!! *****)

open Principal
open Utils

type context = principal list


let void = 
  function () -> []

let add =
  function (ct,np) -> np::ct

let ( ++ ) =
  function ct -> 
    function (mp,np) ->
      let rec sublist = function (ls,el,tls) ->
	match ls with
	| [] -> (tls,[])
	| x::xs when (x=el) -> (tls,xs)
	| x::xs -> sublist(xs,el,(x::tls))
      in
      let (head, tail) = sublist(ct,mp,[]) in
      (np::head)@(tail)

let rec getprincipal =
  function (ct,k) ->
    match ct with
    | x::xs -> if k = 1 then x else getprincipal(xs,k-1)

let getlist =
  function ct -> ct

let setlist =
  function ct -> ct

let getPrev =
  function ct -> List.tl ct

let isEmpty =
  function ct -> forall ct (function x -> Principal.isConsumed x)

let hasEmptyPrincipal =
  function ct -> exists ct (function x -> Principal.isConsumed x)

let howmany =
  function ct -> len ct
      
let getindexedprinc =
  function ct ->
    function i ->
      function asg ->
	let [ttt] = filter ct (function (nn,ii,vlist,expr) -> ii=i) in Principal.substitute asg ttt

let getopenprinc =
  function ct ->
    function asg ->
      let ttt = filter ct (function (nn,ii,vlist,expr) -> vlist<>[]) in
      let mmm = map ttt (function p -> Principal.substitute asg p) in mmm

let getIndexesOf =
  function ct ->
    function s ->
      let ttt = filter ct (function (nn,ii,vlist,expr) -> nn=s) in
      map ttt (function (nn,ii,vlist,expr) -> ii)

let rec getOpenVariables =
  function ct ->
    match ct with
    | [] -> (0,[])
    | pp::pps -> 
	let vs = Principal.getVars pp in 
	let (nn,vv) = getOpenVariables pps in
	(nn+ len(vs),vs@vv)


let getlentrace =
  function ct ->
    if ct = [] then 0
    else fold ct (function a -> function b -> Principal.len a + b) 0


let applyall =
  function ct ->
    function asg ->
      map ct (Principal.substitute asg )

let rec prnt =
  function ct ->
    function asg ->
      match ct with
      | [] -> print_string("\n")
      | x::xs -> Principal.prnt x asg; print_string("\n"); prnt xs asg

let rec prntHtml =
  function ct ->
    function asg ->
      match ct with
      | [] -> print_string("<br>")
      | x::xs -> Principal.prntHtml x asg; print_string("<br>"); prntHtml xs asg

let rec tostring =
  function ct ->
    function asg ->
      match ct with
      | [] -> "\\n"
      | x::xs -> (Principal.tostring x asg)^"\\n"^(tostring xs asg)

