(*
	parsecmdline.ml
	25/05/2003
	Giacomo Baldi

	Implementation of command line parsing functions
*)


type graphopt = {
	mutable context: bool;
	mutable actions: bool;
	mutable k: bool;
	mutable asg: bool;
        mutable level: int;
}

type options = {
    mutable output_type: int;			(* -T, -G: 0 = traces, 1 = Gviz , 2 = NoOutput *)
    mutable attack_type: int;                   (* 0: ascii, 1: latex 2:html*)  
    mutable pr_file: string;			(* needed *)
    mutable pl_file: string;			(* needed *)
    mutable pj_file: string;                    (* needed *)
    mutable kb_file: string;			(* needed *)
    mutable trace_terminal: bool;		(* --terminal | -t: default false *)
    mutable go: graphopt;						(* -Gcontext -Gknowledge -Gassignment -Gactions *)
    mutable errmsg: string;
    mutable help: bool;
    mutable max_princ: int; 				(* -max num: maximun number of principals in context *)
    mutable interactive: bool;
    mutable feedback: bool;
    mutable lazyness: bool;
    mutable mga:bool;
    mutable ic: int list;
    mutable greed: bool;
  }




let init_options = 
  function () -> 
    let t = { output_type = -1; 
	      attack_type = 0;
	      pr_file = ""; 
	      pl_file = ""; 
	      pj_file = "";
	      kb_file = ""; 
	      trace_terminal = false; 
	      go = {
	      actions = false; 
	      k = false; 
	      asg = false; 
	      context = false;
	      level = 16384;
	    }; 
	      errmsg = ""; 
	      help = false; 
	      max_princ = 0; 
	      interactive = false; 
	      feedback = false; 
	      lazyness = true;
	      mga = false;
	      ic = [];
	      greed = false;
	    } in t


let initgo =
  function n ->
    function lev ->
    {
     context = (n >= 8);
     actions = (n mod 8 >= 4);
     k = (n mod 4 >= 2 );
     asg = (n mod 2 = 1);
     level = lev;
   }
      

let parsecmdline =
  function argv ->
    function opts ->
      let i = ref(1) in
      let n = Array.length argv in
      let rec parse =
	function () ->
	  if !i<n then (
	    match argv.(!i) with
	    | "-T" -> (opts.output_type <- 0; i:= !i+1; parse())
	    | "-G" -> 
		opts.output_type <- 1;
		if (!i+1<n) then (
		  try (
		    let mm = int_of_string(argv.(!i+1)) in (opts.go <- initgo mm opts.go.level; i:=!i+2; parse())
		   ) with _ -> (opts.errmsg <- "-G requires numeric option")
		 )
		else (opts.errmsg <- "-G requires numeric option")
	    | "-l" -> 
		if (!i+1<n) then (
		  try (
		    let mm = int_of_string(argv.(!i+1)) in (opts.go.level <-  mm; i:=!i+2; parse())
		   ) with _ -> (opts.errmsg <- "-l requires numeric option")
		 )
		else (opts.errmsg <- "-G requires numeric option")
	    | "-V" -> (opts.output_type <- 2; i:= !i+1; parse())
	    | "-L" -> (opts.attack_type <- 1; i:= !i+1; parse())
	    | "-H" -> (opts.attack_type <- 2; i:= !i+1; parse()) 		  
	    | "-h"
	    | "--help" -> (opts.help <- true; i:= !i+1; parse())
	    | "-lazy" -> (opts.lazyness <- false; i:= !i+1; parse())
	    | "-mga" -> (opts.mga <- true; i:= !i+1; parse())
	    | "-t"
	    | "--terminal" -> (opts.trace_terminal <- true; i:= !i+1; parse())
	    | "--interactive"
	    | "-i"  -> (opts.interactive <- true; i:= !i+1; parse())
	    | "--feedback"
	    | "-f"  -> (opts.feedback <- true; i:= !i+1; parse())
	    | "-g" 
	    | "--greed" ->(opts.greed <- true; i:= !i+1; parse())

	    | "-s" 
	    | "--session"->
		if (!i+1<n) then (
		  let rec roles = 
		    function () ->		 
		      try (
			let mm = int_of_string(argv.(!i+1)) in (opts.ic <- mm::(opts.ic); i:=!i+1);
			roles();
		       ) with _ -> i:=!i+1
		  in
		  roles(); parse();
		  )
		else (opts.errmsg <- "number of roles in a session required")
		


	    | "-max" 
	    | "-nprinc" ->
		if (!i+1<n) then (
		  try (
		    let mm = int_of_string(argv.(!i+1)) in (opts.max_princ <- mm ; i:=!i+2; parse())
		   ) with _ -> (opts.errmsg <- "number of principal in a session required")
		 )
		else (opts.errmsg <- "number of principal in a session required")
	    | s -> 
		if (opts.pr_file = "") then  (opts.pr_file <- s; i:=!i+1; parse())
		else 
		  if (opts.pl_file = "") then  (opts.pl_file <- s; i:=!i+1; parse())
		  else 
		    if (opts.pj_file = "") then  (opts.pj_file <- s; i:=!i+1; parse())
		    else 
		      if (opts.kb_file = "") then  (opts.kb_file <- s; i:=!i+1; parse())
		      else (opts.errmsg <- "don't know what to do with "^s)
			  
	   ) else ()
      in parse()




let synopsis =
  function flag ->
    print_string("\nUsage: tool principals logic join knowledge  [-GVT] [-max nprinc] [-if]\n\n");
    if flag then (
      print_string("\n   -T: Output is text representing the set of traces");
      print_string("\n   -G param: Output is a GraphViz compliant file representing state space tree");
      print_string("\n      param: a numeric value [0..15] specifing what to print. See manual");
      print_string("\n   -V: Enable verification of formula. Output is the description of attacks");
      
      print_string("\n   -t, --terminal: with -T prints out terminal traces only");
      print_string("\n   -i, --interactive: with -V asks for continuation after the discovery of an attack");
      print_string("\n   -f, --feedback: prints on stderr info on search status");

      print_string("\n   -h, --help: print this help");

      print_string("\n")
     ) else ()


let htmlinit =
  function () ->
    print_string("<html><body>")
let htmldone =
  function () ->
    print_string("</body></html>")
