(*
	csolver.ml
	27/07/2003
	Giacomo Baldi

	Constraint Solver for pl verification

*)

open Utils
open Message


type cns = PC of (msg*msg) | NC of (msg*msg) type cs = cns list list

let init =
  function () -> [[]]


let expandable =
  function pc1 ->
    function pc2 ->
      match pc1,pc2 with
      | PC(v1,m1), PC(v2,m2) when m2=v1 -> true
      | PC(v1,m1), NC(v2,m2) when m2=v1 -> true
      | NC(v1,m1), PC(v2,m2) when m2=v1 -> true
      | NC(v1,m1), NC(v2,m2) when m2=v1 -> true
      | _ , _ -> false

let expand =
  function pcte ->
    function pc ->
      match pcte, pc with
      | PC(v1,m1), PC(v2,m2) when v2=m1 -> PC(v1,m2)
      | PC(v1,m1), NC(v2,m2) when v2=m1 -> NC(v1,m2)
      | NC(v1,m1), PC(v2,m2) when v2=m1 -> NC(v1,m2)
      | NC(v1,m1), NC(v2,m2) when v2=m1 -> PC(v1,m2)
      | _ , _ -> pcte


let rec closure =
  function cse ->
    function pc ->
      let ltr = filter cse (expandable pc) in
      let rtl = filter cse ((invert expandable) pc) in
      let toadd1 = map ltr (function x -> expand x pc) in
      let toadd2 = map rtl (function x -> expand pc x) in
      toadd1@toadd2


let rec addc =
  function cse ->
    function pcc ->
      match pcc with
      | PC(v1,m1) when v1=m1 -> cse
      | NC(v1,m1) when v1=m1 -> cse
      | _ ->
	  if exists cse (function x -> x=pcc) then cse
	  else
	    let toadd = closure cse pcc in
	    if toadd = [] then (pcc::cse)
	    else fold toadd (invert (addc) ) (pcc::cse)


let addconstraint =
  function pc ->
    function cs ->
      map cs ((invert addc) pc)

let addconstraints =
  function pcl ->
    function cs ->
      if pcl = [] then cs else fold pcl (addconstraint) cs

let adddisjunct =
  function cs ->
    function pcl ->
      mul cs pcl (addc)

let adddisjuncts =
  function pcl ->
    function cs ->
      if pcl = [] then cs else fold pcl (invert adddisjunct) cs
(*      if pcl = [] then cs 
      else 
	let ncs = debox (grow cs (len pcl)) in
	fold pcl (addcostraints) ncs
*)	  

let addconjuncts =
  function pcl ->
    function cs ->
      if pcl = [] then cs else 
      debox(mul [cs] pcl (adddisjunct))
(*      if pcl = [] then cs 
      else 
	let ncs = debox (grow cs (len pcl)) in
	fold pcl (addcostraints) ncs
*)	  


let toasg =
  function cs ->
    map cs (
    function cse ->
      let pcs = filter cse (function x -> match x with | PC(_,_) -> true | _ -> false) in
      let table = map pcs (function PC(a,b) ->  (a,b)) in
      Assignment.newasg table
   )


let preserve_truth =
  function cs ->
    let rec checkcse =
      function cse ->
	function cso ->
	  match cse with
	  | PC(v,m)::css ->
	      if exists cso (function x -> x=NC(v,m)) then false
	      else 
		if exists cso (
		  function x -> 
		    match x with 
		    | PC(vv,mm) -> v=vv && mm<>m && (notSVar m && notSVar mm && notVar m && notVar mm) 
		    | _ -> false ) 
		then false
		else checkcse css cso
	  | NC(v,m)::css ->
	      if exists cso (function x -> x=PC(v,m)) then false
 	      else checkcse css cso
	  | [] -> true
    in
    filter cs (half (checkcse))

let check =
	function cs ->
		cs = []

let positivize =
  function pcl ->
    map pcl (function x -> PC(x))

let negativize =
  function pcl ->
    map pcl (function x -> NC(x))


let linearize =
  function cs ->
    map (cs) (function x -> [x])


let sss =
function l1 ->
function l2 ->
let cff = 
function v1 ->
function v2 ->
match v1,v2 with
| Var(nv1,(np1,ii1)), Var(nv2,(np2,ii2)) -> (string_of_int(ii1)^nv1) > (string_of_int(ii2)^nv2)
| Var(nv1,(np1,ii1)), SVar((nv2,(np2,ii2)),_,_) -> (string_of_int(ii1)^nv1) > (string_of_int(ii2)^nv2)
| SVar((nv1,(np1,ii1)),_,_), Var(nv2,(np2,ii2)) -> (string_of_int(ii1)^nv1) > (string_of_int(ii2)^nv2)
| SVar((nv1,(np1,ii1)),_,_), SVar((nv2,(np2,ii2)),_,_) -> (string_of_int(ii1)^nv1) > (string_of_int(ii2)^nv2)

in

match l1, l2 with
| PC(v1,m1),PC(v2,m2) -> cff v1 v2
| PC(v1,m1),NC(v2,m2) -> true
| NC(v1,m1),PC(v2,m2) -> false
| NC(v1,m1),NC(v2,m2) -> cff v1 v2

let prnt =
  function cs ->
    print_list cs (
    function cse ->
	     let scse = sort cse (sss) in
      print_list scse (
      function cc ->
	match cc with
(*	| PC(_,SVar(_,_,ty)) -> ()
	| NC(_,SVar(_,_,ty)) -> ()
*)
	| PC(Var(v),m1) -> prntMsg (Var(v)); print_string(" -> "); prntMsg m1; print_string("\n")
	| NC(Var(v),m1) -> prntMsg (Var(v)); print_string(" \\> "); prntMsg m1; print_string("\n")

	| PC(n1,m1) -> prntMsg (n1); print_string(" -> "); prntMsg m1; print_string("\n")
	| NC(n1,m1) -> prntMsg (n1); print_string(" \\> "); prntMsg m1; print_string("\n")

	| _ -> ()
     ); print_string("\n")
   )

let prntHtml = 
  function cs ->
    print_list cs (
    function cse ->
	     let scse = sort cse (sss) in
      print_list scse (
      function cc ->
	match cc with
(*	| PC(_,SVar(_,_,ty)) -> ()
	| NC(_,SVar(_,_,ty)) -> ()
*)
	| PC(Var(v),m1) -> prntMsgHtml (Var(v)); print_string(" -&gt "); prntMsgHtml m1; print_string("<br>")
	| NC(Var(v),m1) -> prntMsgHtml (Var(v)); print_string(" \\&gt "); prntMsgHtml m1; print_string("<br>")

	| PC(n1,m1) -> prntMsgHtml (n1); print_string(" -&gt "); prntMsgHtml m1; print_string("<br>")
	| NC(n1,m1) -> prntMsgHtml (n1); print_string(" \\&gt "); prntMsgHtml m1; print_string("<br>")

	| _ -> ()
     ); print_string("<br>")
   )
