(*
	gviz.ml
	01/05/2003
	Giacomo Baldi

	Definition of functions to format output for GraphViz
*)

open Node
open Parsecmdline


let initgraph =
  function s ->
    print_string("\n/*automatically generated*/\ndigraph "^s^"{\nconcentrate=true\n ")

let donegraph =
  function () ->
    print_string("\n}\n/*end of graph*/\n")

let rec prntTrace =
  function n ->
    function go -> 
      let (ct,asg,kb) = Node.getstatus n in
      let id = Node.getid n in
      let fth = Node.getfather n in
      let level = Node.getlevel n in
      let nodeinfo =
	function () ->
	  print_string("\n");
	  print_string(string_of_int(id)^"[label=\"");

	  if go.actions then (
	    print_string("Action: ");
	    let act = Node.getlabel n in print_string((Action.tostring act)^"\\n\\n");
	   ) else print_string(" ");

	  if go.context then (
	    print_string("Context:\\n------------\\n");
	    print_string(Context.tostring ct asg);
	   ) else ();

	  if go.k then (
	    print_string("Knowledge:\\n------------\\n");
	    print_string(Knowledge.tostring kb);
	   ) else ();

	  if go.asg then (
	    print_string("Assignment:\\n------------\\n");
	    print_string(Assignment.tostring asg);
	   ) else ();

	  print_string("\"]\n");
	  print_string(string_of_int(id)^" ");
      in
      (
       if Node.isNul fth then (
	 print_string(string_of_int(id));
	 nodeinfo();
	 print_string(string_of_int(id)^" ")
	)
       else
	 (
	  prntTrace fth go;
	  if (level<go.level ) then (
	    print_string("-> "^string_of_int(id));
	    nodeinfo()
	   )
	  else ()

	 )
      )


let rec prntTrace2 =
  function n ->
    function bm ->
      function go ->
	let (ct,asg,kb) = Node.getstatus n in
	let id = Node.getid n in
	let fth = Node.getfather n in
	let nodeinfo =
	  function () ->
	    print_string("\n");
	    print_string(string_of_int(id)^"[label=\"");

	    if go.context then (
	      print_string("Context:\\n------------\\n");
	      print_string(Context.tostring ct asg);
	     ) else ();

	    if go.k then (
	      print_string("Knowledge:\\n------------\\n");
	      print_string(Knowledge.tostring kb);
	     ) else ();

	    if go.asg then (
	      print_string("Assignment:\\n------------\\n");
	      print_string(Assignment.tostring asg);
	     ) else ();

	    print_string("\"]\n");
	    print_string(string_of_int(id)^" ");
	in	
	(
	 if Node.isNul fth then (
	   print_string(string_of_int(id));
	   nodeinfo();
	   print_string(string_of_int(id)^" ")
	  )
	 else
	   (
	    let fid = (Node.getid fth) in (
	    (
	     if (bm.(fid)) then print_string(string_of_int(fid)^" ")
	     else prntTrace2 fth bm go
	    );

	    print_string("-> "^string_of_int(id));
	    bm.(fid) <- true;

	    if go.actions then (
	      print_string("[label=\"");
	      let act = Node.getlabel n in print_string(Action.tostring act);
	      print_string("\"] ")
	     ) else print_string(" ");

	    nodeinfo()
	   )
	   )
	)

let rec prntTree =
  function nls ->
    function nn ->
      function go ->
	let bitmap = Array.make nn (false) in
	let rec prntTree2 =
	  function nl ->
	    match nl with
	    | n::ns -> prntTrace2 n bitmap go; print_string("\n"); prntTree2 ns
	    | [] -> print_string("\n")
	in prntTree2 nls

