(*
	Join.ml
	26/05/2003
	Giacomo Baldi

	implementation of module Join
*)

open Node
open Message
open Assignment
open Action
open Princpool
open Utils




let rec iter =
  function i ->
    function j ->
      function f ->
	function en ->
	  if i>j then en
	  else  (f i) (iter (i+1) j f en)

let rec sum =
  function i ->
    function j ->
      function f ->
	iter i j ( compose ( f ) ( + ) ) 0

let rec cardfk =
  function n ->
    function k->
      function i ->
	if n=2 then 1
	else sum 0 (k-1) (cardfk (n-1) (k-i))


let rec getinitialnodes =
  function n ->
    function k->
      if n=2 then k+1
      else sum 0 k (getinitialnodes (n-1))
	  

let initjoin =
  function pp ->
    function kman ->
      function thelist ->
	function node ->
	  let nj = Node.getnj node in
	  let k = Princpool.getmax pp in
	  let rec adder =
	    function nn ->
	      function cl ->
		function ii ->
		  match cl with
		  | x::xs -> (
		      let (oldct, oldasg, oldkb) = Node.getstatus nn in
		      let p = Princpool.getprinc pp (x-1) ii in
		      let name = Principal.getName p in
		      let newkb = (Kmanager.expandK(kman, oldkb, Assignment.void(), Couple(PN(fst(name),snd(name)),PNp(fst(name),snd(name))) )) in
		      let newct = Context.add(oldct,p) in
		      let newst = Status.create(newct, oldasg, newkb) in
		      if ii=1 then (
			let newnode = Node.create(newst) in (
			Node.setfather(newnode, node);
			Node.setlabel(newnode, Join(name, Assignment.void()));
			adder newnode xs (ii+1)
		       )
		       )
		      else (
			Node.setstatus(nn,newst);
			adder nn xs (ii+1)
		       )
		     )
		  | [] -> nn
	  in
	  let rec decodenj =
	    function a ->
	      function b ->
		let c = sum 0 (k) (cardfk (b) (k)) in
		if a-c <0 then (b, a) else decodenj (a-c) (b+1)
	  in
	  let rec gen =
	    function a ->
	      function l ->
		if l = 0 then []
		else a::(gen a (l-1))
	  in
	  let rec generatelist =
	    function npp ->
	      function kk ->
		if npp = 1 then (
		  [gen 1 kk]
		 )
		else (
		  if npp = 2 then (
		    let tmp = ref(-1) in
		    let base = gen 0 (kk+1) in
		    let ttt= map base (function x -> (tmp:=!tmp+1; !tmp)) in
		    map ttt (function x -> (gen 1 (kk-x))@(gen 2 x))
		   )
		  else
		    let part = iter 0 kk (compose (generatelist (npp-1)) ( @ )) [] in
		    map part (
		    function l ->
		      let sz = len l in
		      l@(gen npp (kk-sz))
		   )
		 )
	  in
	  let rec onecnt =
	    function l ->
	      function ix ->
	      match l with
	      | r::rs -> (gen ix r)@(onecnt rs (ix+1))
	      | [] -> []				     
	  in
	  if nj= -1 then Node.nulNode()
	  else
	    let (oldct, oldasg, oldkb) = Node.getstatus node in
	    if (Context.howmany oldct=k) then (
	      Node.setnj(node,-1);
	      Node.nulNode()
	     ) else (
	      if (thelist = []) then ( 	      
		let codelist = (generatelist (howmany pp) k) in
		let flag = (nj >= (len codelist)) in
		if flag then (
		  Node.setnj(node,-1);
		  Node.nulNode()
		 ) else (
		  let sel = ( select codelist nj) in (
		  (*
		     print_list sel (function s -> print_string(string_of_int(s)^"."));
		     print_string(" --> "^string_of_int(nj)^"\n");
		   *)
		  Node.setnj(node,nj+1);
		  adder node sel 1
		 )
		 )
	       ) else (
		Node.setnj(node,-1);
		adder node (onecnt thelist 1) 1;
	       )
	     )

let rec newAsg = 
  function ls ->
    function ks ->
      match ls with
      | Var(v)::vs -> Assignment.add (newAsg vs ks) (Var(v)) (K("k!"^ks,0))
      | [] -> Assignment.void()
	    
    

      
let joincontext =
  function kman -> 
    function greed ->
      function node ->
	let (ct,asg,kb) = Node.getstatus node in
	let newlct = Context.getlist (Context.applyall ct asg) in 
	let kbid = Knowledge.id_of kb in
	let rec closeprincipal =
	  function pp ->
	    match pp with
	    | (a,b,v::vs,c) ->
		let tasg = 
		  if (Principal.typeOf pp v) <> SKey then (Assignment.add (Assignment.void()) (Var(v)) (SVar(v,kbid,Name)))
		  else Assignment.void()
		in tasg*(closeprincipal (a,b,vs,c))
	    | (_,_,[],_) -> (Assignment.void())
	in
	let rec closep =
	  function prs ->
	    match prs with
	    | pp::pps -> 
		let newasg = closeprincipal pp in
		newasg*(closep pps)
	    | [] -> (Assignment.void())
	in 
	let rec closeprincipalgreed =
	  function pp ->
	    match pp with
	    | (a,b,v::vs,c) ->
		let tasg = 
		  if (Principal.typeOf pp v) <> SKey then (
		    let names = Knowledge.getAllNames kb in
		    map names (function n -> (Assignment.add (Assignment.void()) (Var(v)) (n)))
		   )
		  else [Assignment.void()]
		in mul tasg (closeprincipalgreed (a,b,vs,c)) (Assignment.comp)
	    | (_,_,[],_) -> [(Assignment.void())]
	in
	let rec closepgreed =
	  function prs ->
	    match prs with
	    | pp::pps -> 
		let newasg = closeprincipalgreed pp in
		mul newasg (closepgreed pps) (Assignment.comp)
	    | [] -> [Assignment.void()]
	in 

	let rec instantiateKeys =
	  function part ->
	    match part with
	    | x::xs ->
		let nk = ref(0) in
		let nasg = fold x (function ls -> function fasg -> nk:= !nk+1; ((newAsg ls (string_of_int(!nk)))*fasg)) (Assignment.void()) in
		nasg::(instantiateKeys xs )
	    | [] -> []
	in
	let ctasg = 
	  if (greed) then (
	    let tempasg = (closepgreed newlct) in
	    let newctg = Context.applyall ct (head tempasg) in
 	    let (nk,vars) = (Context.getOpenVariables newctg) in
	    let parts = partition (vars@(Knowledge.getIntruderVariables kb)) in
	    let asgres = instantiateKeys parts in
	    let ctasgg = if (asgres = []) then [Assignment.void()] else asgres in	
	    mul tempasg ctasgg (Assignment.comp)	  
	   )
	  else(
	    let tempasg =  (closep newlct) in
	    let newct = Context.applyall ct tempasg in
	    let (nk,vars) = (Context.getOpenVariables newct) in
	    let parts = partition (vars@(Knowledge.getIntruderVariables kb)) in
	    let asgres = instantiateKeys parts in
	    if (asgres = []) then [tempasg] else map asgres (Assignment.comp tempasg)
	      
	   )
	in
	let allnewct = map ctasg (function sigma -> Context.applyall ct (asg*sigma)) in
	let allnewkb = map ctasg (function sigma -> Kmanager.updateK (kman, kb, sigma)) in
	let allnewst = couplify (couplify ctasg allnewct) allnewkb in
	let allnewnode = map allnewst (
	  function ((newasg,newct),newkb) -> 
	    let newnode = Node.create(Status.create(newct, asg*newasg, newkb) ) in (
	    Node.setfather (newnode, node);
	    Node.setlabel (newnode, Join(("Close Principals",Node.getid newnode),asg*newasg));
	    newnode
	   )
	 ) in allnewnode
	  
	  
let joinwithmodel =
  function kman ->
    function greed ->
    function node ->
      function mdls ->
	if map mdls (fst) = [] then ([])
	else 
	  let (ct,asg,kb) = Node.getstatus node in
	  debox (
	  map mdls (
	  function (ss,cs) ->
	    let newasg = asg*ss in
	    let newst = Status.create(ct, newasg, kb) in
	    let newnode = Node.create(newst) in (
	    Node.setfather (newnode, node);
	    Node.setlabel (newnode, Join(("Check Component",Node.getid newnode),ss));
	    map (joincontext kman greed newnode) (function nnn -> (nnn,cs))
	   )	      
	 )
	 )

	    

