(*
	process.mli
	23/01/2003
	Giacomo Baldi
	
	Interface for process.ml
	Definition of process representation with abstract syntax tree
*)

(**
   This module implements an internal representation od the process of a principal. It is based on cIP a restriction
   of pi-calculus.
*)


open Assignment
open Message

(**
   when the result of a substitution is a semantically wrong process
*)
exception SubstitutionException
	
(**
   A process is used to represent principal actions.
*)
type process = 
  | In of msg*process (** Input action: principal is trying to input a message m. (in(k).out(... ) *)
  | Out of msg*process (** Output action: principal is trying to output a message m, (out(k).in(... ) *)
  | Sum of process*process (** Non determinism: every choiche of action is possible. in(k) + out(n) *)
  | Par of process*process (** Parallelism: two actions can be made at the same time. in(k) | out(n) *)
  | Nil (** No action *)
      
      
(**
   substitute takes an assignment asg and a process p.
   @return a new process np obtained from p apply asg to every message in p.
   @raise SubstitutionException when the result is malformed
*)
val substitute: Assignment.assignment -> process -> process

(**
   validate takes a process p.
   @return true if p i semantically correct, false otherwise.
*)
val validate: process -> bool

(**
   len takes a process p.
   @return the max number of input/output actions in p.
*)
val len: process -> int
	    
(**
   The following functions are used to print a representation of a process.
*)

val prnt: process -> unit
val prntHtml: process -> unit

val tostring: process -> string
	    
