open Message
open Logic
(* An indexlist is a list of indexes, initialized as empty list reference*)
let indexlist = ref ([])

(* indices is a function that populates the indexlist with all indices of the formula passed to this function as input*)

let rec indices phi indexlist = match phi with 
  |Forall(i,s,psi)->  indexlist:=i::!indexlist; (indices psi indexlist)
  |Exists(i,s,psi)->  indexlist:=i::!indexlist ;  (indices psi indexlist)
  |Not(psi1)->indices psi1 indexlist
  |And(psi1,psi2)->(indices psi1 indexlist) ;(indices psi2 indexlist)
  |Or(psi1,psi2)-> (indices psi1 indexlist) ; (indices psi2 indexlist)
  |_->indexlist

(*Freshindex function generates a freshindex and append it to the indexlist*)


(*Repindex is a function that replaces an oldindex with a newindex in a Message *)
let rec freshindex i indexlist =
 if   List.mem i !indexlist then 
(let newindex = i^string_of_int (Random.int 15) in freshindex newindex indexlist)
 else
(indexlist:=i::!indexlist;i)

let rec repindex m oldi newi = match m with  
  |K(s,i) -> if  i=  0 then  
 	(let j= String.index s 'k' in
	 K("_"^newi^String.sub s j ((String.length s)- (j)),0))
	else  K(s,i) 
|NO(s,i)-> if  i=  0   then
	(let j= String.index s 'n' in NO("_"^newi^String.sub s j ((String.length s)- (j)),0))
 	 else  NO(s,i) 
|PN(s,i)->  if i = 0   then 
    		(let j = try (String.index s oldi.[0]) with Not_found -> -1 in 
 			 if j = -1 then  m  
  			else 
  			PN((String.sub s 0 j)^newi,0))
 	else m
|PNp(s,i)-> if i = 0  then
 		(let j = try (String.index s oldi.[0]) with Not_found -> -1 in 
  			if j = -1 then   m  
  			else 
  			PNp((String.sub s 0 j)^newi,0))
 	 else m 
|PNm(s,i)-> if i = 0 then
		(let j = try (String.index s oldi.[0]) with Not_found -> -1 in 
  			if j = -1  then   m  
  			else 
  	  		PNm((String.sub s 0 j)^newi,0))
	    else m 
|Var(n,(i,m)) when i=oldi-> Var(n,(newi,m))
|Publ(Var(n,(i,m))) when i=oldi-> Publ(Var(n,(newi,m)))
|Priv(Var(n,(i,m))) when i=oldi-> Priv(Var(n,(newi,m)))
|Crypt(m1,m2) ->Crypt(repindex m1 oldi newi, repindex m2 oldi newi)
|Couple(m1,m2)->Couple(repindex m1 oldi newi, repindex m2 oldi newi)
|_->m 




(*replace function recieves as input the old and new index and a formula and replace the old one with new one in the formula*)
let rec replace i1 i2 phi = 
match phi with 
  |Forall(i,s,psi)->Forall(i,s, replace i1  i2 psi)
  |Exists(i,s,psi)->Exists(i,s,replace i1 i2 psi)
  |Not(psi)->Not(replace i1 i2 psi)
  |And(psi1,psi2)->And(replace i1 i2 psi1, replace i1 i2 psi2) 
  |Or(psi1,psi2)->Or(replace i1 i2 psi1, replace i1 i2 psi2) 
  |Equal(m1,m2)->Equal(repindex m1 i1 i2,repindex m2 i1 i2)
  |Diff(m1,m2)->Diff(repindex m1 i1 i2,repindex m2 i1 i2)
  |Derive(m)->Derive(repindex m i1 i2)
  |NDerive(m)->NDerive(repindex m i1 i2)
  |True->True
  |False->False;;


(* pnf is a function that accepts a PL formula and returns it's equivelant PNF. 
If the formula strats with Forall/Exists its is already in PNF and then we check for psi for PNF.
If the formula starts with Not then we check the subformula for any quantifier (if the subformula has a qunatifier we swap Forall with Exists and Exists for Forall and recursively check the remaining formula)
If the formula starts with AND or Or we check the two subformulas for its PNF, if there are quantifiers in the subformula , they can be brought outside in front of the formula with the condition that we have to replace the indices with new index
*)


let rec pnf phi =  match phi with 
  |Forall(i,s,psi)-> Forall(i,s,pnf psi)
  |Exists(i,s,psi)->Exists(i,s,pnf psi)
  |Not(psi1)->  (match pnf (psi1) with 
		  |Forall(i,s,psi2)->Exists(i,s,pnf (Not(psi2)))
		  |Exists(i,s,psi2)->Forall(i,s,pnf (Not(psi2)))
		  |_->Not(psi1))
  |And(psi1,psi2)->( match pnf (psi1) with
    		  |Forall(i,s,psi1')-> 
			let newi = freshindex i (indices phi indexlist) in 
			Forall(newi,s,pnf (And((replace i newi psi1'),psi2)))
      		|Exists(i,s,psi1')->
			let newi = freshindex i (indices phi indexlist) in 
			Exists(newi,s,pnf (And((replace i newi psi1'),psi2)))
    		|_-> ( match pnf (psi2) with
     		 |Forall(i,s,psi2')-> 
			let newi = freshindex i (indices phi indexlist) in 
			Forall(newi,s,pnf (And(psi1,(replace i newi psi2'))))
    		|Exists(i,s,psi2')->
			let newi = freshindex i (indices phi indexlist) in 
			Exists(newi,s,pnf (And(psi1,(replace i newi psi2'))))
     		|_->And(psi1,psi2)))
  |Or(psi1,psi2)->( match pnf (psi1) with
    		  |Forall(i,s,psi1')-> 
			let newi = freshindex i (indices phi indexlist) in 
			Forall(newi,s,pnf (Or((replace i newi psi1'),psi2)))
    		  |Exists(i,s,psi1')->
			let newi = freshindex i (indices phi indexlist) in 
			Exists(newi,s,pnf (Or((replace i newi psi1'),psi2)))
     		 |_-> ( match pnf (psi2) with
   		   |Forall(i,s,psi2')-> 
			let newi = freshindex i (indices phi indexlist) in 
			Forall(newi,s,pnf (Or(psi1,(replace i newi psi2'))))
     		  |Exists(i,s,psi2')->
			let newi = freshindex i (indices phi indexlist) in 
			Exists(newi,s,pnf (Or(psi1,(replace i newi psi2'))))
   		   |_->Or(psi1,psi2)))
  |_->phi
