(*
	verifier.ml
	25/07/2003
	Giacomo Baldi

	Implementation of pl logic verifier

*)
open Logic
open Assignment
open Core
open Utils
open Csolver
open Message

type ops = Positive | Negative | Noop


let solve_membership =
  function Derive(m) ->
    function kman ->
      function node ->
	let (ct, asg, k) = Node.getstatus node in
	map (mu_eq kman k asg m) (snd)

let solve_equality =
  function Equal(v,m) ->
    function kman ->
      function node ->
	let (ct, asg, k) = Node.getstatus node in
	ni_eq kman asg v m


(*Controllare il caso di assegnamento vuoto*)
let rec solveconjunct =
  function pls ->
    function kman ->
      function node ->
	match pls with
	| True::lls -> 
	    let (a,b,c,d) = solveconjunct lls kman node in
	    (True::a, True::b, c, d)	      
	| False::lls -> 
	    let (a,b,c,d) = solveconjunct lls kman node in
	    (False::a, False::b, c, d)
	| Equal(v,m)::lls ->
	    let asgl = solve_equality (Equal(v,m)) kman node in
	    if asgl = [] then (False::lls, False::lls, [], Positive)
	    else if (Assignment.isVoid (head asgl)) then (True::lls, True::lls, [], Positive) 
	    else (True::lls, False::lls, asgl, Positive)
	| Diff(v,m)::lls ->
	    let asgl = solve_equality (Equal(v,m)) kman node in
	    if asgl = [] then (True::lls, True::lls, [], Positive)
	    else if (Assignment.isVoid (head asgl)) then (False::lls, False::lls, [], Positive)
	    else (True::lls, False::lls, asgl, Negative)
	| Derive(m)::lls ->
	    let asgl = solve_membership (Derive(m)) kman node in
	    if asgl = [] then (False::lls, False::lls, [], Positive)
	    else if (Assignment.isVoid (head asgl)) then (True::lls, True::lls, [], Positive) 
	    else (True::lls, False::lls, asgl, Positive)
	| NDerive(m)::lls ->
	    let asgl = solve_membership (Derive(m)) kman node in
	    if asgl = [] then (True::lls, True::lls, [], Positive)
	    else if (Assignment.isVoid (head asgl)) then (False::lls, False::lls, [], Positive) 
	    else (True::lls, False::lls, asgl, Negative)
	| [] -> ([],[],[], Noop)


let rec solve1 =
  function pl ->
    function kman ->
      function node ->
	match pl with
	| t::ts ->
	    let (a,b,c,d) = solveconjunct t kman node in
	    if (d=Noop) then let (e,f,g,h) = solve1 ts kman node in (a::e,b::f,g,h)
	    else (a::ts, b::ts, c, d)
	| [] -> ([], [], [], Noop)

	      
	      

let solve =
  function stack ->
    function kman ->
      function node ->
	let (pl,pasg,css) = Thestack.top stack in
	let cs = preserve_truth css in
	if (check cs) then (([], pasg, cs),Thestack.pop stack)
	else  (
	  (*
	     print_string("\nNow processing:\n");
	  Logic.prntnpl pl;
	  flush stdout;
	     *)
	  let (newpl1, newpl2, asg, tops) = solve1 pl kman node in

	  if (tops=Noop) then (
	    (*verification terminated*)
	    ((newpl1, pasg, cs),(Thestack.pop stack))
	   )
	  else (
	    if (asg=[]) then (
	      (*Unification failed -> unicity of result*) 
	      (([],pasg,cs), Thestack.push (Thestack.pop stack) ((newpl1, pasg, cs)))
	     ) else (
	      (*New constraints -> double result*)
	      if tops = Positive then (
		let newpstack = map asg (
		  function x ->
		    let lll = positivize(listfy x) in
		    (concretize newpl1 x, pasg*x, addconstraints lll cs)
		 )
		in
		let newfstack = (newpl2, pasg, fold asg (
				 function x ->
				   function y ->
				     let lll = negativize(listfy x) in
				     (addconstraints lll y)
				) cs)
		in
		let newstack = newfstack::newpstack in
		(([],pasg,cs),Thestack.pushl (Thestack.pop stack) newstack)
	       )
	      else (
		let newpstack = map asg (
		  function x ->
		    let lll = positivize(listfy x) in
		    (concretize newpl2 x, pasg*x, addconstraints lll cs)
		 )
		in
		let newfstack = (newpl1, pasg, fold asg (
				 function x ->
				   function y ->
				     let lll = negativize(listfy x) in
				     (addconstraints lll y)
				) cs)
		in
		let newstack = newfstack::newpstack in
		(([],pasg,cs),Thestack.pushl (Thestack.pop stack) newstack)
	       )
	     )
	   )	    
	 )

(************************)

let solver = 
  function bottom ->
    function kman ->
      function node ->
	let pstack = ref (Thestack.init bottom) in
	let result = ref ([]) in
	(
	 while (not (Thestack.isEmpty (!pstack))) do
	   (
    	    (*prerr_string("solve "^string_of_int(Thestack.length !pstack)^"\n"); flush stderr;*)
	    let (tres, tstack) = (solve (!pstack) kman node) in
    	    (*prerr_string("!solve\n"); flush stderr;*)
	    pstack := tstack;
	    match tres with
	    | ([],_,_) -> ()
	    | _ -> result := tres::(!result)
	   )
	 done;
	 !result
	)



let expandasg =
  function asg ->
    function kman ->
      let lcs = (listfy asg)  in
      let kx = Kmanager.getK kman 1 in
      if lcs = [] then []
      else
	let toexp = fold lcs (
	  function (a,b) ->
	    function ls ->
	      let t=asg a in
	      (
	       match a,t with
	       | Var(_), SVar(_,_,ty) when ((ty<>Gen && ty<> GenR))->
		   let res = mu_eq kman kx asg t in
		   (positivize(map res (function (mm,ss) -> (t,mm))))::ls
	       | SVar(v,_,ty), _ when ((ty = Gen || ty = GenR) && notSVar(t)) -> 
		   (positivize([(a,t)]))::ls
	       | _  -> ls
	      )
	 ) [] in
	(*(map lcs (function x -> positivize ([x])))@*)remclone toexp ( <> )



let checkpl =
  function pl ->
    function css ->
      function kman ->
	function node ->
	  function vr ->
	    let verify =
	      function pl ->
		let ppp = map pl (function x -> if exists x (function a -> a=False) then False else True) in
		if exists ppp (function x -> x=True) then True else False
	    in
	    let (ct,asg,k) = Node.getstatus node in
	    let npl = concretize pl asg in
	    let lcs = positivize (listfy asg) in
	    let csf = Csolver.adddisjuncts (expandasg asg kman) (css)  in

(*
   print_string("ah!\n");
   Logic.prntnpl pl;
   Csolver.prnt csf;
 *)
(*	    print_list (expandasg asg kman) (
	    function cse ->
	      print_list cse (
	      function cc ->
		match cc with
		| PC(v,m1) -> prntMsg (v); print_string(" -> "); prntMsg m1; print_string("\n")
		| NC(v,m1) -> prntMsg (v); print_string(" \\> "); prntMsg m1; print_string("\n")
		| _ -> ()
	     ); print_string("||\n")
	   ); flush stdout;
*)
 	    let fff = solver ((npl, asg, csf)) kman node in
(*	let ttt = map fff (function (fr, asgf, cssf) -> (verify fr, asgf, cssf)) in *)
	    filter fff (function (fr,asgf,cssf) -> (verify fr = vr) && (not (Csolver.check cssf)))


	      (*filter fff (function (f,a) -> forall f (function x -> x=[False]))*)

let checkplf = 
  function pl ->
    function css ->
    function kman ->
      function node -> checkpl pl css kman node False

let checkplt = 
  function pl ->
    function css ->
    function kman ->
      function node -> checkpl pl css kman node True


let getmodels =
  function pl ->
    function kman ->
      function node ->
	if (forall pl (function cnj -> exists cnj (function x -> x=False))) then []
	else
	  let res = checkplt pl (Csolver.init()) kman node in
	  (

(*	 print_string("--------------\n"); 
   Logic.prntnpl pl;
 *)
	   let asgs = debox(map res (
			    function (fr, asgf, cssf) -> 

(*
   print_string("model\n");
   Csolver.prnt cssf;
   
 *)
			      map (Csolver.toasg cssf) (function ss -> (ss,cssf)))) in
	   asgs
	  ) 

let getemptymodel =
  function () -> (Assignment.void(),Csolver.init())



(* fast verifier *)


let fast_solveconjunct =
  function pls ->
    function kman ->
      function node ->
	let rec fast_solve_positive_atoms =
	  function spls ->
	    match spls with  
	    | Equal(v,m)::lls ->
		let asgl = solve_equality (Equal(v,m)) kman node in
		if asgl = [] then (False::lls, [], Positive)
		else (True::lls, asgl, Positive)
	    | Derive(m)::lls ->
		let asgl = solve_membership (Derive(m)) kman node in
		if asgl = [] then (False::lls, [], Positive)
		else (True::lls, asgl, Positive)
	    | ll::lls ->
  		let (a,c,d) = fast_solve_positive_atoms lls in
		(ll::a, c, d)	      
	    | [] -> ([],[], Noop)
	in
	let rec fast_solve_negative_atoms =
	  function spls ->
	    match spls with
	    | Diff(v,m)::lls ->
		let asgl = solve_equality (Equal(v,m)) kman node in
		if asgl = [] then (True::lls, [], Negative)
		else 
		  if (Assignment.isVoid (head asgl)) then (False::lls, [], Negative)
		  else (True::lls, asgl, Negative)
	    | NDerive(m)::lls ->
		let asgl = solve_membership (Derive(m)) kman node in
		if asgl = [] then (True::lls, [], Negative)
		else 
		  if (Assignment.isVoid (head asgl)) then (False::lls, [], Negative)
		  else (True::lls, asgl, Negative)
	    | ll::lls ->
  		let (a,c,d) = fast_solve_negative_atoms lls in
		(ll::a, c, d)	      
	    | [] -> ([],[], Noop)
	in
	
	let pres = (fast_solve_positive_atoms pls) in 
	match pres  with
	| (_,_,Noop) -> fast_solve_negative_atoms pls 
	| _ -> pres
	      

let fast_solve =
  function stack ->
    function kman ->
      function node ->
	let (pl,pasg,css) = Thestack.top stack in
	let cs = preserve_truth css in
	if (check cs) then (([], pasg, cs),Thestack.pop stack)
	else  (
	  let (newpl1, asg, tops) = fast_solveconjunct pl kman node in
	  if (tops=Noop) then (
	    (*verification terminated*)
	    ((newpl1, pasg, cs),(Thestack.pop stack))
	   )
	  else (
	    if (asg=[]) then (
	      (*Unification failed -> unicity of result*)
	      (([],pasg,cs), Thestack.push (Thestack.pop stack) ((newpl1, pasg, cs)))
	     ) else (
	      (*New constraints -> concretizing results*) 
	      if (tops=Positive) then (
		(*concretizing with assignments*)
		let newpstack = map asg (
		  function x ->
		    let lll = positivize(listfy x) in
		    (debox (concretize ([newpl1]) x), pasg*x, addconstraints lll cs)
		 )
		in (([],pasg,cs), Thestack.pushl (Thestack.pop stack) newpstack)
	       ) else (
		let newfstack = (newpl1, pasg, (addconjuncts (map asg (function x -> negativize (listfy x))) cs))
				
		in (([],pasg,cs), Thestack.push (Thestack.pop stack) newfstack)
	       )
	     )
	   )	    
	 )


let fast_solver = 
  function bottom ->
    function kman ->
      function node ->
	let pstack = ref (Thestack.init bottom) in
	let result = ref ([]) in
	(
	 while (not (Thestack.isEmpty (!pstack))) do
	   (

(*    	    prerr_string("fast_solve "^string_of_int(Thestack.length !pstack)^"\n"); flush stderr;*)

	    let (tres, tstack) = (fast_solve (!pstack) kman node) in
	    pstack := tstack;
	    match tres with
	    | ([],_,_) -> ()
	    | _ -> result := (tres)::(!result)
	   )
	 done;
	 !result
	)



let fast_checkpl =
  function pl ->
    function css ->
      function kman ->
	function node ->
	  let (ct,asg,k) = Node.getstatus node in
	  let npl = concretize pl asg in
	  let lcs = positivize (listfy asg) in
	  let csf = Csolver.adddisjuncts (expandasg asg kman) css (*(Csolver.addconstraints lcs css)*)  in
	  let rec scan_conjuncts =
	    function ipl ->
	      function pll ->
	      match ipl,pll with
	      | ipls::iplls, pls::plls ->
		  let res = fast_solver ((pls,asg,csf)) kman node in
		  let ok = filter res (function (lp,lasg,lcsf) -> (forall lp (function x -> (x=True))) && (not (Csolver.check csf))) in
		  if (ok = []) then scan_conjuncts iplls plls else (ipls,map ok (function (lp,lasg,lcsf) -> ([lp],lasg,lcsf)))
	      | [],[] -> ([],[])
	  in
	  scan_conjuncts pl npl


let very_fast_checkpl =
  function pl ->
    function css ->
      function kman ->
	function node ->
	  let (ct,asg,k) = Node.getstatus node in
	  let npl = concretize pl asg in
	  let csf = Csolver.adddisjuncts (expandasg asg kman) css in 
	  let rec scan_conjuncts =
	    function ipl ->
	      function pll ->
	      match ipl,pll with
	      | ipls::iplls, pls::plls ->
		  let res = fast_solver ((pls,asg,csf)) kman node in
		  let ok = filter res (function (lp,lasg,lcsf) -> (forall lp (function x -> (x=True))) && (not (Csolver.check csf))) in
		  if (ok = []) then scan_conjuncts iplls plls else (ipls,map ok (function (lp,lasg,lcsf) -> ([lp],lasg,lcsf)))
	      | [],[] -> ([],[])
	  in
	  scan_conjuncts pl npl



let fast_sat_model =
  function pl ->
    function cs ->
      function kman ->
	function node ->
	  let res = very_fast_checkpl pl (cs) kman node in
	  (res <> ([],[]) )


let fast_sat = 
  function pl ->
    function kman ->
      function node ->
	fast_sat_model pl (Csolver.init()) kman node


